from typing import Callable, Optional, Union

import webviz_core_components as wcc
import webviz_subsurface_components as wsc
from dash import dcc, html

from .._utils import datetime_utils
from ..figures.color_figure import color_figure
from ..models import ParametersModel, SimulationTimeSeriesModel


def ensemble_selector(
    get_uuid: Callable,
    ensembles: list,
    tab: str,
    id_string: str,
    multi: bool = False,
    value: str = None,
    heading: str = None,
) -> html.Div:
    return wcc.Dropdown(
        label=heading,
        id={"id": get_uuid(id_string), "tab": tab},
        options=[{"label": ens, "value": ens} for ens in ensembles],
        multi=multi,
        value=value if value is not None else ensembles[0],
        clearable=False,
    )


def vector_selector(get_uuid: Callable, vectormodel: SimulationTimeSeriesModel):
    return wsc.VectorSelector(
        id=get_uuid("vector-selector"),
        maxNumSelectedNodes=1,
        data=vectormodel.vector_selector_data,
        persistence=True,
        persistence_type="session",
        selectedTags=["FOPT"]
        if "FOPT" in vectormodel.vectors
        else vectormodel.vectors[:1],
        numSecondsUntilSuggestionsAreShown=0.5,
        lineBreakAfterTag=True,
    )


def parameter_selector(
    get_uuid: Callable, parametermodel: ParametersModel, tab: str
) -> html.Div:
    return wcc.Dropdown(
        label="Parameter",
        id={"id": get_uuid("parameter-select"), "tab": tab},
        options=[{"label": i, "value": i} for i in parametermodel.parameters],
        placeholder="Select a parameter...",
        clearable=False,
    )


def sortby_selector(
    get_uuid: Callable,
    value: str = None,
) -> html.Div:
    return html.Div(
        wcc.RadioItems(
            label="Sort parameters by",
            id=get_uuid("delta-sort"),
            className="block-options",
            options=[
                {"label": "Name", "value": "Name"},
                {
                    "label": "Standard deviation",
                    "value": "Stddev",
                },
                {
                    "label": "Average",
                    "value": "Avg",
                },
            ],
            value=value,
        )
    )


def plot_options(get_uuid: Callable, tab: str) -> html.Div:
    return html.Div(
        children=[
            wcc.Checklist(
                id={"id": get_uuid("checkbox-options"), "tab": tab},
                options=[
                    {"label": "Dateline visible", "value": "DateLine"},
                    {"label": "Auto compute correlations", "value": "AutoCompute"},
                ],
                value=["DateLine", "AutoCompute"],
            ),
            dcc.Store(
                id={"id": get_uuid("plot-options"), "tab": tab}, storage_type="session"
            ),
        ],
    )


def date_selector(
    get_uuid: Callable, vectormodel: SimulationTimeSeriesModel
) -> html.Div:
    dates = vectormodel.dates
    return html.Div(
        style={"margin": "10px 0px"},
        children=[
            html.Div(
                style={"display": "inline-flex"},
                children=[
                    wcc.Label("Date:"),
                    wcc.Label(
                        datetime_utils.to_str(dates[-1]),
                        id=get_uuid("date-selected-text"),
                        style={"margin-left": "10px"},
                    ),
                    dcc.Store(
                        id=get_uuid("date-selected"),
                        storage_type="session",
                        data=datetime_utils.to_str(dates[-1]),
                    ),
                ],
            ),
            wcc.Slider(
                id=get_uuid("date-slider"),
                value=len(dates) - 1,
                min=0,
                max=len(dates) - 1,
                step=1,
                included=False,
                marks={
                    idx: {
                        "label": datetime_utils.to_str(dates[idx]),
                        "style": {"white-space": "nowrap"},
                    }
                    for idx in [0, len(dates) - 1]
                },
            ),
        ],
    )


def filter_parameter(
    get_uuid: Callable,
    parametermodel: ParametersModel,
    tab: str,
    multi: bool = True,
    value: Union[str, float] = None,
) -> html.Div:
    return html.Div(
        style={"margin-top": "10px"},
        children=[
            wcc.Label("Select parameters"),
            html.Div(
                id=get_uuid("filter-parameter-container"),
                children=[
                    wcc.SelectWithLabel(
                        id={
                            "id": get_uuid("filter-parameter"),
                            "tab": tab,
                        },
                        options=[
                            {"label": i, "value": i} for i in parametermodel.parameters
                        ],
                        value=value,
                        multi=multi,
                        size=min(40, len(parametermodel.parameters)),
                    ),
                ],
            ),
        ],
    )


def make_filter(
    get_uuid: Callable,
    tab: str,
    vtype: str,
    column_values: list,
    multi: bool = True,
    value: Union[str, float] = None,
    open_details: bool = False,
) -> wcc.Selectors:
    return html.Details(
        open=open_details,
        children=[
            html.Summary(vtype),
            wcc.Select(
                id={
                    "id": get_uuid("vitem-filter"),
                    "tab": tab,
                    "vtype": vtype,
                },
                options=[{"label": i, "value": i} for i in column_values],
                value=[value] if value is not None else column_values,
                multi=multi,
                size=min(15, len(column_values)),
            ),
        ],
    )


def button(uuid, label, height="30px"):
    return html.Button(
        label,
        id=uuid,
        style={
            "marginBottom": "5px",
            "width": "100%",
            "height": height,
            "line-height": height,
            "background-color": "#E8E8E8",
        },
    )


def filter_vector_selector(get_uuid: Callable) -> html.Div:
    return html.Div(
        children=[
            dcc.Textarea(
                id=get_uuid("vector-filter"),
                style={"width": "95%", "height": "60px", "resize": "none"},
                placeholder="\nenter comma separated input\n* is used as wildcard",
                persistence=True,
                persistence_type="session",
            ),
            button(get_uuid("submit-vector-filter"), "Submit", height="20px"),
            dcc.Store(id=get_uuid("vector-filter-store")),
        ]
    )


def color_selector(
    get_uuid: Callable,
    tab: str,
    colors: Optional[list] = None,
    bargap: Optional[float] = None,
    height: Optional[float] = None,
):
    return html.Div(
        style={"margin-top": "5px"},
        children=[
            wcc.Label("Colors"),
            wcc.Graph(
                style={"height": height},
                id={"id": get_uuid("color-selector"), "tab": tab},
                config={"displayModeBar": False},
                figure=color_figure(
                    colors=colors,
                    bargap=bargap,
                    height=height,
                ),
            ),
        ],
    )


def color_opacity_selector(get_uuid: Callable, tab: str, value: float):
    return html.Div(
        style={"margin-top": "5px"},
        children=[
            "Opacity:",
            dcc.Input(
                id={"id": get_uuid("opacity-selector"), "tab": tab},
                type="number",
                min=0,
                max=1,
                step=0.1,
                value=value,
                style={"margin-left": "10px"},
            ),
        ],
    )



from pathlib import Path
from typing import Callable, Dict, List, Tuple
from uuid import uuid4

import pandas as pd
import webviz_subsurface_components as wsc
from dash import Dash, Input, Output, State, callback_context, dcc, html
from dash.exceptions import PreventUpdate
from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.common_cache import CACHE
from webviz_config.webviz_store import webvizstore

from .._datainput.fmu_input import load_parameters


class ParameterDistribution(WebvizPluginABC):
    """Visualizes parameter distributions for FMU ensembles.

Parameters are visualized either as histograms, showing parameter ranges
and distributions for each ensemble.

Input can be given either as an aggregated `csv` file with parameter information,
or as ensemble name(s) defined in `shared_settings`.

---

**Using aggregated data**
* **`csvfile`:** Aggregated `csv` file with `REAL`, `ENSEMBLE` and parameter columns. \
 (absolute path or relative to config file).

**Reading data from ensembles**
* **`ensembles`:** Which ensembles in `shared_settings` to visualize.

---
Parameter values are extracted automatically from the `parameters.txt` files in the individual
realizations if you have defined `ensembles`, using the `fmu-ensemble` library.

When using an aggregated `csvfile`, you need to have the columns `REAL`, `ENSEMBLE`
and the parameter columns.
"""

    def __init__(
        self,
        app: Dash,
        webviz_settings: WebvizSettings,
        csvfile: Path = None,
        ensembles: list = None,
    ):

        super().__init__()

        self.csvfile = csvfile if csvfile else None

        if csvfile and ensembles:
            raise ValueError(
                'Incorrect arguments. Either provide a "csvfile" or "ensembles".'
            )
        if csvfile:
            self.parameters = read_csv(csvfile)
        elif ensembles:
            self.ensembles = {
                ens: webviz_settings.shared_settings["scratch_ensembles"][ens]
                for ens in ensembles
            }
            self.parameters = load_parameters(
                ensemble_paths=self.ensembles, ensemble_set_name="EnsembleSet"
            )
        else:
            raise ValueError(
                'Incorrect arguments. Either provide a "csvfile" or "ensembles".'
            )

        self.parameter_columns: List[str] = [
            col
            for col in list(self.parameters.columns)
            if col not in ["REAL", "ENSEMBLE"]
        ]
        self.uid = uuid4()
        self.plotly_theme = webviz_settings.theme.plotly_theme
        self.set_callbacks(app)

    def ids(self, element: str) -> str:
        """Generate unique id for dom element"""
        return f"{element}-id-{self.uid}"

    @property
    def tour_steps(self) -> List[dict]:
        return [
            {
                "id": self.ids("layout"),
                "content": ("Dashboard displaying distribution of input parameters"),
            },
            {
                "id": self.ids("graph"),
                "content": (
                    "Visualization of currently selected parameter as histogram "
                    "series and distribution range per ensemble."
                ),
            },
            {
                "id": self.ids("parameter"),
                "content": (
                    "Select visualized parameter by selecting or searching the list."
                ),
            },
        ]

    @staticmethod
    def set_grid_layout(columns: str) -> Dict[str, str]:
        return {
            "display": "grid",
            "alignContent": "space-around",
            "justifyContent": "space-between",
            "gridTemplateColumns": f"{columns}",
        }

    def make_buttons(self, prev_id: str, next_id: str) -> html.Div:
        return html.Div(
            style=self.set_grid_layout("1fr 1fr"),
            children=[
                html.Button(
                    id=prev_id,
                    style={
                        "fontSize": "2rem",
                        "paddingLeft": "5px",
                        "paddingRight": "5px",
                    },
                    children="⬅",
                ),
                html.Button(
                    id=next_id,
                    style={
                        "fontSize": "2rem",
                        "paddingLeft": "5px",
                        "paddingRight": "5px",
                    },
                    children="➡",
                ),
            ],
        )

    @property
    def layout(self) -> html.Div:
        return html.Div(
            id=self.ids("layout"),
            children=[
                html.Span("Parameter distribution:", style={"font-weight": "bold"}),
                html.Div(
                    style=self.set_grid_layout("8fr 1fr 2fr"),
                    children=[
                        dcc.Dropdown(
                            id=self.ids("parameter"),
                            options=[
                                {"value": col, "label": col}
                                for col in self.parameter_columns
                            ],
                            value=self.parameter_columns[0],
                            clearable=False,
                            persistence=True,
                            persistence_type="session",
                        ),
                        self.make_buttons(self.ids("prev-btn"), self.ids("next-btn")),
                    ],
                ),
                wsc.PriorPosteriorDistribution(id=self.ids("graph")),
            ],
        )

    def set_callbacks(self, app: Dash) -> None:
        @app.callback(
            Output(self.ids("parameter"), "value"),
            [
                Input(self.ids("prev-btn"), "n_clicks"),
                Input(self.ids("next-btn"), "n_clicks"),
            ],
            [State(self.ids("parameter"), "value")],
        )
        def _set_parameter_from_btn(
            _prev_click: int, _next_click: int, column: str
        ) -> str:

            ctx = callback_context.triggered
            if ctx is None:
                raise PreventUpdate
            callback = ctx[0]["prop_id"]
            if callback == f"{self.ids('prev-btn')}.n_clicks":
                column = prev_value(column, self.parameter_columns)
            elif callback == f"{self.ids('next-btn')}.n_clicks":
                column = next_value(column, self.parameter_columns)
            else:
                column = self.parameter_columns[0]
            return column

        @app.callback(
            Output(self.ids("graph"), "data"), [Input(self.ids("parameter"), "value")]
        )
        def _set_parameter(column: str) -> dict:
            param = self.parameters[[column, "REAL", "ENSEMBLE"]]

            ensembles = param["ENSEMBLE"].unique().tolist()

            iterations = []
            values = []
            labels = []

            for ensemble in ensembles:
                df = param[param["ENSEMBLE"] == ensemble]
                iterations.append(ensemble)
                values.append(df[column].tolist())
                labels.append([f"Realization {real}" for real in df["REAL"].tolist()])

            return {"iterations": iterations, "values": values, "labels": labels}

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        return [
            (read_csv, [{"csv_file": self.csvfile}])
            if self.csvfile
            else (
                load_parameters,
                [
                    {
                        "ensemble_paths": self.ensembles,
                        "ensemble_set_name": "EnsembleSet",
                    }
                ],
            )
        ]


def prev_value(current_value: str, options: List[str]) -> str:
    try:
        index = options.index(current_value)
        return options[max(0, index - 1)]
    except ValueError:
        return current_value


def next_value(current_value: str, options: List[str]) -> str:
    try:
        index = options.index(current_value)
        return options[min(len(options) - 1, index + 1)]
    except ValueError:
        return current_value


@CACHE.memoize(timeout=CACHE.TIMEOUT)
@webvizstore
def read_csv(csv_file: Path) -> pd.DataFrame:
    return pd.read_csv(csv_file, index_col=None)



import datetime
from pathlib import Path
from typing import Optional

import pandas as pd
import pytest
from pandas._testing import assert_frame_equal

from webviz_subsurface._models.gruptree_model import GruptreeModel, TreeType

CHECK_COLUMNS = ["DATE", "CHILD", "KEYWORD", "PARENT"]
ENSEMBLE = "01_drogon_ahm"
GRUPTREE_FILE = "share/results/tables/gruptree.csv"


@pytest.fixture(name="gruptree_model")
def fixture_model(testdata_folder) -> GruptreeModel:
    ens_path = Path(testdata_folder) / ENSEMBLE / "realization-*" / "iter-0"
    return GruptreeModel(
        ens_name="iter-0",
        ens_path=ens_path,
        gruptree_file=GRUPTREE_FILE,
        tree_type="GRUPTREE",
    )


# Mock class that loads local csv file
class MockGruptreeModel(GruptreeModel):
    # pylint: disable=super-init-not-called
    def __init__(self, tree_type: Optional[TreeType] = None):
        self._tree_type = tree_type
        df_files = pd.DataFrame([{"REAL": 0, "FULLPATH": "tests/data/gruptree.csv"}])
        self._dataframe = self.read_ensemble_gruptree(df_files=df_files)


@pytest.mark.usefixtures("app")
def test_gruptree_model_init(testdata_folder, gruptree_model: GruptreeModel):

    # Check that there is only one REAL (means that the gruptree is
    # the same for all realizations)
    assert gruptree_model.dataframe["REAL"].nunique() == 1

    # Load gruptree table from realization-0 and compare with
    # the dataframe from the gruptree_model
    r0_path = f"{testdata_folder}/{ENSEMBLE}/realization-0/iter-0/{GRUPTREE_FILE}"
    exp_df = pd.read_csv(r0_path)
    exp_df["DATE"] = pd.to_datetime(exp_df["DATE"])
    exp_df = exp_df.where(pd.notnull(exp_df), None)

    assert_frame_equal(gruptree_model.dataframe[CHECK_COLUMNS], exp_df[CHECK_COLUMNS])


@pytest.mark.usefixtures("app")
def test_get_filtered_dataframe(gruptree_model: GruptreeModel):

    # Test the get_filtered_dataframe function with terminal node different than FIELD
    filtered_df = gruptree_model.get_filtered_dataframe(terminal_node="OP")
    filtered_df = filtered_df[
        filtered_df["DATE"] == filtered_df["DATE"].max()
    ].reset_index()
    exp_filtered_df = pd.DataFrame(
        columns=["DATE", "CHILD", "KEYWORD", "PARENT"],
        data=[
            [datetime.datetime(year=2018, month=11, day=17), "OP", "GRUPTREE", "FIELD"],
            [datetime.datetime(year=2018, month=11, day=17), "A1", "WELSPECS", "OP"],
            [datetime.datetime(year=2018, month=11, day=17), "A2", "WELSPECS", "OP"],
            [datetime.datetime(year=2018, month=11, day=17), "A3", "WELSPECS", "OP"],
            [datetime.datetime(year=2018, month=11, day=17), "A4", "WELSPECS", "OP"],
        ],
    )
    assert_frame_equal(filtered_df[CHECK_COLUMNS], exp_filtered_df)

    # Test excl_wells_startswith and excl_wells_endswith
    assert set(
        gruptree_model.get_filtered_dataframe(
            excl_well_startswith=["R_"],
            excl_well_endswith=["3", "5"],
        )["CHILD"].unique()
    ) == {"FIELD", "OP", "RFT", "WI", "A1", "A2", "A4", "A6"}


def test_tree_type_filtering():

    mock_model = MockGruptreeModel(tree_type=TreeType.GRUPTREE)
    assert "BRANPROP" not in mock_model.dataframe["KEYWORD"].unique()

    mock_model = MockGruptreeModel(tree_type=TreeType.BRANPROP)
    assert "GRUPTREE" not in mock_model.dataframe["KEYWORD"].unique()

    # If tree_type is defaulted then the BRANPROP tree is selected
    mock_model = MockGruptreeModel()
    assert "GRUPTREE" not in mock_model.dataframe["KEYWORD"].unique()



from pathlib import Path

import pytest
import xtgeo

from webviz_subsurface._datainput.fmu_input import find_surfaces
from webviz_subsurface._models.surface_set_model import SurfaceSetModel


@pytest.mark.usefixtures("app")
def test_surface_set_model(testdata_folder):
    ensemble_paths = {
        "iter-0": str(
            Path(testdata_folder / "01_drogon_ahm" / "realization-*" / "iter-0")
        )
    }

    surface_table = find_surfaces(ensemble_paths)
    surface_table = surface_table.drop("ENSEMBLE", axis=1)

    smodel = SurfaceSetModel(surface_table)
    assert set(smodel.attributes) == set(
        [
            "ds_extract_postprocess",
            "amplitude_mean",
            "ds_extract_geogrid",
            "amplitude_rms",
            "oilthickness",
        ]
    )
    assert set(smodel.names_in_attribute("ds_extract_postprocess")) == set(
        ["basevolantis", "topvolantis", "toptherys", "topvolon"]
    )
    real_surf = smodel.get_realization_surface(
        attribute="ds_extract_postprocess", name="topvolon", realization=0
    )
    assert isinstance(real_surf, xtgeo.RegularSurface)
    assert real_surf.values.mean() == pytest.approx(1735.42, 0.00001)
    stat_surf = smodel.calculate_statistical_surface(
        attribute="ds_extract_postprocess", name="topvolon"
    )
    assert isinstance(stat_surf, xtgeo.RegularSurface)
    assert stat_surf.values.mean() == pytest.approx(1741.04, 0.00001)

    stat_surf = smodel.calculate_statistical_surface(
        attribute="ds_extract_postprocess", name="topvolon", realizations=[0, 1]
    )
    assert stat_surf.values.mean() == pytest.approx(1741.04, 0.00001)



import json
import pathlib
import warnings
from typing import Dict, Optional, Tuple, cast

import pandas as pd
from webviz_subsurface_components import VectorDefinition, VectorDefinitions

_DATA_PATH = pathlib.Path(__file__).parent.absolute() / "abbreviation_data"

SIMULATION_VECTOR_TERMINOLOGY = VectorDefinitions

RESERVOIR_SIMULATION_UNIT_TERMINOLOGY = json.loads(
    (_DATA_PATH / "reservoir_simulation_unit_terminology.json").read_text()
)


def simulation_unit_reformat(ecl_unit: str, unit_set: str = "METRIC") -> str:
    """Returns the simulation unit in a different, more human friendly, format if possible,
    otherwise returns the simulation unit.
    * `ecl_unit`: Reservoir simulation vector unit to reformat
    * `unit_set`: Currently only valid option is the default "METRIC" (defined as in Eclipse E100)
    """
    return RESERVOIR_SIMULATION_UNIT_TERMINOLOGY[unit_set].get(ecl_unit, ecl_unit)


def simulation_vector_base(vector: str) -> str:
    """Returns base name of a simulation vector on Eclipse format.
    E.g. WOPR for WOPR:OP_1 and ROIP for ROIP_REG:1.
    Some description: If a vector contains a :, the first up to 5 characters describe the vector
    base. In that case, if the vector base name is shorter than 5 characters, _ is used to fill
    up the first five characters. Therefore splitting on first _ and limiting to max five
    characters to identify the base name.
    """
    return vector.split(":", 1)[0].split("_", 1)[0][:5] if ":" in vector else vector


def simulation_vector_description(
    vector: str,
    user_defined_vector_definitions: Optional[Dict[str, VectorDefinition]] = None,
) -> str:
    """Returns a more human friendly description of the simulation vector if possible,
    otherwise returns the input as is.

    # TODO: Remove support for "AVG_" and "INTVL_" when all usage is deprecated.
    """
    prefix = ""
    suffix = ""
    if vector.startswith("AVG_"):
        prefix = "Average "
        vector = vector[4:]
    elif vector.startswith("PER_DAY_"):
        prefix = "Average "
        suffix = " Per day"
        vector = vector[8:]
    elif vector.startswith("INTVL_"):
        prefix = "Interval "
        vector = vector[6:]
    elif vector.startswith("PER_INTVL_"):
        prefix = "Interval "
        vector = vector[10:]

    vector_name: str
    node: Optional[str]
    if ":" in vector:
        [vector_name, node] = vector.split(":", 1)
    else:
        vector_name = vector
        node = None

    def _get_vector_definition(vector: str) -> Optional[VectorDefinition]:
        """Get vector definition. Fetch user defined if existing"""
        if (
            user_defined_vector_definitions
            and vector in user_defined_vector_definitions
        ):
            return user_defined_vector_definitions[vector]
        if vector in SIMULATION_VECTOR_TERMINOLOGY:
            return SIMULATION_VECTOR_TERMINOLOGY[vector]
        return None

    if len(vector_name) == 8:
        if vector_name[0] == "R":
            # Region vectors for other FIP regions than FIPNUM are written on a special form:
            # 8 signs, with the last 3 defining the region.
            # E.g.: For an array "FIPREG": ROIP is ROIP_REG, RPR is RPR__REG and ROIPL is ROIPLREG
            # Underscores _ are always used to fill
            [vector_base_name, fip] = [vector_name[0:5].rstrip("_"), vector_name[5:]]

            _definition = _get_vector_definition(vector_base_name)
            if _definition and _definition["type"] == "region":
                return (
                    f"{prefix}{_definition['description']}{suffix}, region {fip} {node}"
                )
        elif vector_name.startswith("W") and vector_name[4] == "L":
            # These are completion vectors, e.g. WWCTL:__1:OP_1 and WOPRL_10:OP_1 for
            # water-cut in OP_1 completion 1 and oil production rate in OP_1 completion 10
            [vector_base_name, comp] = [vector_name[0:5], vector_name[5:].lstrip("_")]

            _definition = _get_vector_definition(vector_base_name)
            if _definition and _definition["type"] == "completion":
                return (
                    f"{prefix}{_definition['description']}"
                    f"{suffix}, well {node} completion {comp}"
                )

    _definition = _get_vector_definition(vector_name)
    if _definition:
        if node is None:
            return prefix + _definition["description"] + suffix
        return (
            f"{prefix}{_definition['description']}{suffix}, "
            f"{_definition['type'].replace('_', ' ')} {node}"
        )

    if not vector.startswith(
        ("AU", "BU", "CU", "FU", "GU", "RU", "SU", "WU", "Recovery Factor of")
    ):
        # Vectors starting with AU, BU, CU, FU, GU, RU, SU and WU are user defined vectors.
        # Currently not providing descriptions for these, but migth come later (see:
        # https://github.com/equinor/webviz-subsurface/issues/321)
        warnings.warn(
            (
                f"Could not find description for vector {vector_name}. Consider adding"
                " it in the GitHub repo https://github.com/equinor/webviz-subsurface-components?"
            ),
            UserWarning,
        )
    return prefix + vector + suffix


def historical_vector(
    vector: str,
    smry_meta: Optional[pd.DataFrame] = None,
    return_historical: Optional[bool] = True,
) -> Optional[str]:
    """This function is trying to make a best guess on converting between historical and
    non-historical vector names.

    `vector`: An Eclipse-format vector string
    `smry_meta`: Note: Not activate avaiting https://github.com/equinor/libecl/issues/708
                 A pandas DataFrame with vector metadata on the format returned by
                 `load_smry_meta` in `../_datainput/fmu_input.py`. Here the field is_historical is
                 used to check if a vector is a historical vector.
    `return_historical`: If return_historical is `True`, the corresponding guessed historical
                         vector name is returned if the guessed vector is thought to be a
                         historical vector, else None is returned. If `False` the corresponding
                         non-historical vector name is returned, if the input vector is thought to
                         be a historical vector, else None is returned.
    """
    smry_meta = None  # Temp deactivation waiting on https://github.com/equinor/libecl/issues/708
    parts = vector.split(":", 1)
    if return_historical:
        parts[0] += "H"
        hist_vec = ":".join(parts)
        return (
            None
            if historical_vector(hist_vec, smry_meta=smry_meta, return_historical=False)
            is None
            else hist_vec
        )

    if smry_meta is None:
        if parts[0].endswith("H") and parts[0].startswith(("F", "G", "W")):
            parts[0] = parts[0][:-1]
            return ":".join(parts)
        return None

    try:
        is_hist = smry_meta.is_historical[vector]
    except KeyError:
        is_hist = False
    return parts[0][:-1] if is_hist else None


def simulation_region_vector_breakdown(
    vector: str,
) -> Tuple[str, Optional[str], Optional[str]]:
    [vector_base_name, node, fip] = _vector_breakdown(vector)
    if fip is not None and len(fip) == 3:
        fiparray = f"FIP{fip}"
    elif fip is not None:
        fiparray = cast(str, fip)
    else:
        fiparray = ""
    return vector_base_name, fiparray, node


def simulation_region_vector_recompose(
    vector_base_name: str, fiparray: str, node: str
) -> str:
    return (
        vector_base_name
        + (
            "_" * (5 - len(vector_base_name)) + fiparray[-3:]
            if fiparray != "FIPNUM"
            else ""
        )
        + f":{node}"
    )


def _vector_breakdown(vector: str) -> Tuple[str, Optional[str], Optional[str]]:
    vector_name: str
    node: Optional[str]
    fip: Optional[str]
    if ":" in vector:
        [vector_name, node] = vector.split(":", 1)
    else:
        vector_name = vector
        node = None
    if len(vector_name) == 8:
        # Region vectors for other FIP regions than FIPNUM are written on a special form:
        # 8 signs, with the last 3 defining the region.
        # E.g.: For an array "FIPREG": ROIP is ROIP_REG, RPR is RPR__REG and ROIPL is ROIPLREG
        # Underscores _ are always used to fill
        [vector_base_name, fip] = [vector_name[0:5].rstrip("_"), vector_name[5:]]
        try:
            if SIMULATION_VECTOR_TERMINOLOGY[vector_base_name]["type"] == "region":
                vector_name = vector_base_name
            else:
                fip = None
        except KeyError:
            fip = None
    else:
        try:
            fip = (
                "NUM"
                if SIMULATION_VECTOR_TERMINOLOGY[vector_name]["type"] == "region"
                else "FIELD"
                if SIMULATION_VECTOR_TERMINOLOGY[vector_name]["type"] == "field"
                else None
            )

        except KeyError:
            fip = None
    return vector_name, node, fip



# pylint: disable=no-name-in-module
from webviz_config.plugins import WellLogViewer


def test_well_log_viewer(dash_duo, app, testdata_folder) -> None:
    wellfolder = testdata_folder / "observed_data" / "wells/"
    plugin = WellLogViewer(
        app,
        wellfolder=wellfolder,
        wellsuffix=".rmswell",
        mdlog="MDepth",
        logtemplates=[f"{testdata_folder}/webviz_examples/all_logs_template.yml"],
    )
    app.layout = plugin.layout
    dash_duo.start_server(app)
    assert dash_duo.get_logs() == []



from typing import Dict, List, Optional

from webviz_subsurface._providers import VectorMetadata

from ....mocks.ensemble_summary_provider_dummy import EnsembleSummaryProviderDummy

##################################################################################################
#
# Mock implementations of EnsembleSummaryProvider for testing of EnsembleSummaryProviderSet class
#
# Only methods utilized by the EnsembleSummaryProviderSet is implemented for the
# EnsembleSummaryProvider mock
#
##################################################################################################


class EnsembleSummaryProviderMock(EnsembleSummaryProviderDummy):
    """Class for EnsembleSummaryProvider mock implementations for testing of
    EnsembleSummaryProviderSet class.

    Contains implementation of methods for utilize in unit tests.
    """

    def __init__(
        self,
        dataset_name: str,
        vector_metadata_dict: Dict[str, VectorMetadata],
        realizations: List[int],
    ) -> None:
        self._vector_metadata_dict: Dict[
            str, VectorMetadata
        ] = vector_metadata_dict.copy()
        self._vector_names: List[str] = list(self._vector_metadata_dict.keys())
        self._realizations: List[int] = realizations
        self._dataset_name = dataset_name

    @staticmethod
    def create_mock_with_first_dataset() -> "EnsembleSummaryProviderMock":
        vector_metadata_dict = {
            "WWCT:A1": VectorMetadata(
                unit="",
                is_total=False,
                is_rate=True,
                is_historical=False,
                keyword="WWCT",
                wgname="A1",
                get_num=6,
            ),
            "WWCT:A2": VectorMetadata(
                unit="",
                is_total=False,
                is_rate=True,
                is_historical=False,
                keyword="WWCT",
                wgname="A2",
                get_num=7,
            ),
            "WGOR:A1": VectorMetadata(
                unit="SM3/SM3",
                is_total=False,
                is_rate=True,
                is_historical=False,
                keyword="WGOR",
                wgname="A1",
                get_num=6,
            ),
            "WGOR:A2": VectorMetadata(
                unit="SM3/SM3",
                is_total=False,
                is_rate=True,
                is_historical=False,
                keyword="WGOR",
                wgname="A2",
                get_num=7,
            ),
            "WBHP:A1": VectorMetadata(
                unit="BARSA",
                is_total=False,
                is_rate=False,
                is_historical=False,
                keyword="WBHP",
                wgname="A1",
                get_num=6,
            ),
            "WBHP:A2": VectorMetadata(
                unit="BARSA",
                is_total=False,
                is_rate=False,
                is_historical=False,
                keyword="WBHP",
                wgname="A2",
                get_num=7,
            ),
        }
        realizations = [1, 2, 3, 4, 5]
        return EnsembleSummaryProviderMock(
            "First dataset", vector_metadata_dict, realizations
        )

    @staticmethod
    def create_mock_with_second_dataset() -> "EnsembleSummaryProviderMock":
        vector_metadata_dict = {
            "WOPT:A1": VectorMetadata(
                unit="SM3",
                is_total=True,
                is_rate=False,
                is_historical=False,
                keyword="WOPT",
                wgname="A1",
                get_num=6,
            ),
            "WOPT:A2": VectorMetadata(
                unit="SM3",
                is_total=False,
                is_rate=True,
                is_historical=False,
                keyword="WOPT",
                wgname="A2",
                get_num=7,
            ),
            "WWCT:A1": VectorMetadata(
                unit="",
                is_total=False,
                is_rate=True,
                is_historical=False,
                keyword="WWCT",
                wgname="A1",
                get_num=6,
            ),
            "WWCT:A2": VectorMetadata(
                unit="",
                is_total=False,
                is_rate=True,
                is_historical=False,
                keyword="WWCT",
                wgname="A2",
                get_num=7,
            ),
            "WOPR:A2": VectorMetadata(
                unit="SM3/DAY",
                is_total=False,
                is_rate=True,
                is_historical=False,
                keyword="WOPR",
                wgname="A2",
                get_num=7,
            ),
            "FGIR": VectorMetadata(
                unit="SM3/DAY",
                is_total=False,
                is_rate=True,
                is_historical=False,
                keyword="FGIR",
                wgname=None,
                get_num=0,
            ),
        }
        realizations = [1, 2, 4, 5, 8]
        return EnsembleSummaryProviderMock(
            "Second dataset", vector_metadata_dict, realizations
        )

    @staticmethod
    def create_mock_with_third_dataset() -> "EnsembleSummaryProviderMock":
        vector_metadata_dict = {
            "WGOR:A1": VectorMetadata(
                unit="SM3/SM3",
                is_total=False,
                is_rate=True,
                is_historical=False,
                keyword="WGOR",
                wgname="A1",
                get_num=6,
            ),
            "WGOR:A2": VectorMetadata(
                unit="SM3/SM3",
                is_total=False,
                is_rate=True,
                is_historical=False,
                keyword="WGOR",
                wgname="A2",
                get_num=7,
            ),
            "WOPR:A1": VectorMetadata(
                unit="SM3/DAY",
                is_total=False,
                is_rate=True,
                is_historical=False,
                keyword="WOPR",
                wgname="A1",
                get_num=6,
            ),
            "WOPR:A2": VectorMetadata(
                unit="SM3/DAY",
                is_total=False,
                is_rate=True,
                is_historical=False,
                keyword="WOPR",
                wgname="A2",
                get_num=7,
            ),
            "FGIR": VectorMetadata(
                unit="SM3/DAY",
                is_total=False,
                is_rate=True,
                is_historical=False,
                keyword="FGIR",
                wgname=None,
                get_num=0,
            ),
            "FGIT": VectorMetadata(
                unit="SM3",
                is_total=True,
                is_rate=False,
                is_historical=False,
                keyword="FGIT",
                wgname=None,
                get_num=0,
            ),
        }
        realizations = [1, 2, 3, 4, 7]
        return EnsembleSummaryProviderMock(
            "Third dataset", vector_metadata_dict, realizations
        )

    @staticmethod
    def create_mock_with_inconsistent_dataset() -> "EnsembleSummaryProviderMock":
        """Mock implementation to define inconsistent metadata

        Introduces metadata info breaking with the metadata in the three previous implementations
        """
        vector_metadata_dict = {
            "WWCT:A1": VectorMetadata(
                unit="Invalid Unit",
                is_total=False,
                is_rate=False,
                is_historical=False,
                keyword="WWCT",
                wgname="A1",
                get_num=6,
            ),
            "WWCT:A2": VectorMetadata(
                unit="Invalid Unit",
                is_total=False,
                is_rate=False,
                is_historical=False,
                keyword="WWCT",
                wgname="A2",
                get_num=7,
            ),
            "WGOR:A1": VectorMetadata(
                unit="SM3",
                is_total=False,
                is_rate=False,
                is_historical=True,
                keyword="WGOR",
                wgname="A1",
                get_num=6,
            ),
            "WGOR:A2": VectorMetadata(
                unit="SM3",
                is_total=False,
                is_rate=False,
                is_historical=True,
                keyword="WGOR",
                wgname="A2",
                get_num=7,
            ),
        }
        realizations = [1, 2, 4, 5, 6]
        return EnsembleSummaryProviderMock(
            "Inconsistent", vector_metadata_dict, realizations
        )

    def get_dataset_name(self) -> str:
        return self._dataset_name

    ########################################
    #
    # Override methods
    #
    ########################################
    def vector_names(self) -> List[str]:
        return self._vector_names

    def realizations(self) -> List[int]:
        return self._realizations

    def vector_metadata(self, vector_name: str) -> Optional[VectorMetadata]:
        return self._vector_metadata_dict.get(vector_name, None)



from typing import Any, List

import webviz_core_components as wcc
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC

from ...._types import VfpParam


class Vizualisation(SettingsGroupABC):
    class Ids(StrEnum):
        COLOR_BY = "color-by"

    def __init__(self, vfp_names: List[str]) -> None:
        super().__init__("Vizualisation")
        self._vfp_names = vfp_names

    def layout(self) -> List[Any]:
        return [
            wcc.Dropdown(
                id=self.register_component_unique_id(Vizualisation.Ids.COLOR_BY),
                label="Color by",
                options=[
                    {"label": "THP", "value": VfpParam.THP},
                    {"label": "WFR", "value": VfpParam.WFR},
                    {"label": "GFR", "value": VfpParam.GFR},
                    {"label": "ALQ", "value": VfpParam.ALQ},
                ],
                clearable=False,
                value=VfpParam.THP,
                persistence=True,
                persistence_type="session",
            )
        ]



import json
import sys
from dataclasses import dataclass
from typing import Any, Dict, List, Tuple, Union

import pandas as pd
from dash import Input, Output, callback
from dash.exceptions import PreventUpdate
from webviz_config.utils import StrEnum, callback_typecheck
from webviz_config.webviz_plugin_subclasses import ViewABC

from webviz_subsurface._components.tornado._tornado_data import TornadoData
from webviz_subsurface._components.tornado._tornado_table import TornadoTable

from ...shared_settings import FilterOption, Scale
from .view_elements import TornadoTable as TornadoTableViewElement

if sys.version_info >= (3, 9):
    from typing import Annotated
else:
    from typing_extensions import Annotated


class TornadoTableView(ViewABC):
    class IDs(StrEnum):
        TORNADO_TABLE = "tornado-table"
        MAIN_COLUMN = "main-column"

    @dataclass
    class Slots:
        reference: Annotated[Input, str]
        scale: Annotated[Input, str]
        filter_options: Annotated[Input, List[str]]
        data: Annotated[Input, Union[str, bytes, bytearray]]
        sens_filter: Annotated[Input, Union[List[str], str]]

    def __init__(self, design_matrix_df: pd.DataFrame, slots: Slots) -> None:
        super().__init__("Table View")

        self._design_matrix_df = design_matrix_df
        self._slots = slots

        column = self.add_column(TornadoTableView.IDs.MAIN_COLUMN)
        first_row = column.make_row()
        first_row.add_view_element(
            TornadoTableViewElement(), TornadoTableView.IDs.TORNADO_TABLE
        )

    def set_callbacks(self) -> None:
        @callback(
            Output(
                self.view_element(TornadoTableView.IDs.TORNADO_TABLE)
                .component_unique_id(TornadoTableViewElement.IDs.TABLE)
                .to_string(),
                "data",
            ),
            Output(
                self.view_element(TornadoTableView.IDs.TORNADO_TABLE)
                .component_unique_id(TornadoTableViewElement.IDs.TABLE)
                .to_string(),
                "columns",
            ),
            self._slots.reference,
            self._slots.scale,
            self._slots.filter_options,
            self._slots.data,
            self._slots.sens_filter,
        )
        @callback_typecheck
        def _calc_tornado_table(
            reference: str,
            scale: Scale,
            filter_options: List[FilterOption],
            data: Union[str, bytes, bytearray],
            sens_filter: Union[List[str], str],
        ) -> Tuple[List[Any], List[Dict[Any, Any]]]:
            if not data:
                raise PreventUpdate
            filter_options = filter_options if filter_options else []
            data = json.loads(data)
            if not isinstance(sens_filter, List):
                sens_filter = [sens_filter]
            if not isinstance(data, dict):
                raise PreventUpdate
            values = pd.DataFrame(data["data"], columns=["REAL", "VALUE"])
            realizations = self._design_matrix_df.loc[
                self._design_matrix_df["ENSEMBLE"] == data["ENSEMBLE"]
            ]

            design_and_responses = pd.merge(values, realizations, on="REAL")
            if sens_filter is not None:
                if reference not in sens_filter:
                    sens_filter.append(reference)
                design_and_responses = design_and_responses.loc[
                    design_and_responses["SENSNAME"].isin(sens_filter)
                ]
            tornado_data = TornadoData(
                dframe=design_and_responses,
                response_name=data.get("response_name"),
                reference=reference,
                scale="Percentage" if scale == Scale.REL_VALUE_PERC else "Absolute",
                cutbyref=FilterOption.REMOVE_SENS_WITH_NO_IMPACT in filter_options,
            )
            tornado_table = TornadoTable(tornado_data=tornado_data)
            return (
                tornado_table.as_plotly_table,
                tornado_table.columns,
            )



# pylint: disable=no-name-in-module
from webviz_config.plugins import PvtPlot
from webviz_config.testing import WebvizComposite


def test_pvt_plot(_webviz_duo: WebvizComposite, shared_settings: dict) -> None:
    plugin = PvtPlot(
        shared_settings["HM_SETTINGS"],
        ensembles=shared_settings["HM_ENSEMBLES"],
        pvt_relative_file_path="share/results/tables/pvt.csv",
    )

    _webviz_duo.start_server(plugin)

    assert _webviz_duo.get_logs() == []



from typing import Callable

from dash import dcc, html


def clientside_stores(get_uuid: Callable) -> html.Div:
    """Contains the clientside stores"""
    return html.Div(
        children=[
            dcc.Store(id=get_uuid("selections"), storage_type="session"),
            dcc.Store(id=get_uuid("page-selected"), storage_type="session"),
            dcc.Store(id=get_uuid("voldist-page-selected"), storage_type="session"),
            dcc.Store(id=get_uuid("initial-load-info"), storage_type="memory"),
            html.Div(
                style={"display": "none"},
                children=dcc.Download(id=get_uuid("download-dataframe")),
            ),
        ]
    )



import numpy as np
import pyarrow as pa

from webviz_subsurface._providers.ensemble_summary_provider._resampling import (
    Frequency,
    generate_normalized_sample_dates,
    interpolate_backfill,
    sample_segmented_multi_real_table_at_date,
)


def _create_table_from_row_data(
    per_row_input_data: list, schema: pa.Schema
) -> pa.Table:
    # Turn rows into columns
    columns_with_header = list(zip(*per_row_input_data))

    input_dict = {}
    for col in columns_with_header:
        colname = col[0]
        coldata = col[1:]
        input_dict[colname] = coldata

    table = pa.Table.from_pydict(input_dict, schema=schema)

    return table


def test_generate_sample_dates_daily() -> None:

    dates = generate_normalized_sample_dates(
        np.datetime64("2020-12-30"), np.datetime64("2021-01-05"), Frequency.DAILY
    )
    assert len(dates) == 7
    assert dates[0] == np.datetime64("2020-12-30")
    assert dates[-1] == np.datetime64("2021-01-05")

    dates = generate_normalized_sample_dates(
        np.datetime64("2020-12-30T01:30"),
        np.datetime64("2021-01-05T02:30"),
        Frequency.DAILY,
    )
    assert len(dates) == 8
    assert dates[0] == np.datetime64("2020-12-30")
    assert dates[-1] == np.datetime64("2021-01-06")


def test_generate_sample_dates_weekly() -> None:

    # Mondays
    #   2020-12-21
    #   2020-12-28
    #   2021-01-04
    #   2021-01-11

    dates = generate_normalized_sample_dates(
        np.datetime64("2020-12-28"), np.datetime64("2021-01-11"), Frequency.WEEKLY
    )
    assert len(dates) == 3
    assert dates[0] == np.datetime64("2020-12-28")
    assert dates[-1] == np.datetime64("2021-01-11")

    dates = generate_normalized_sample_dates(
        np.datetime64("2020-12-27T00:01"),
        np.datetime64("2021-01-05T02:30"),
        Frequency.WEEKLY,
    )
    assert len(dates) == 4
    assert dates[0] == np.datetime64("2020-12-21")
    assert dates[-1] == np.datetime64("2021-01-11")


def test_generate_sample_dates_monthly() -> None:

    dates = generate_normalized_sample_dates(
        np.datetime64("2020-12-01"), np.datetime64("2021-01-01"), Frequency.MONTHLY
    )
    assert len(dates) == 2
    assert dates[0] == np.datetime64("2020-12-01")
    assert dates[-1] == np.datetime64("2021-01-01")

    dates = generate_normalized_sample_dates(
        np.datetime64("2020-12-30"),
        np.datetime64("2022-01-01T01:01"),
        Frequency.MONTHLY,
    )
    assert len(dates) == 15
    assert dates[0] == np.datetime64("2020-12-01")
    assert dates[-1] == np.datetime64("2022-02-01")


def test_generate_sample_dates_yearly() -> None:

    dates = generate_normalized_sample_dates(
        np.datetime64("2020-01-01"), np.datetime64("2020-01-02"), Frequency.YEARLY
    )
    assert len(dates) == 2
    assert dates[0] == np.datetime64("2020-01-01")
    assert dates[-1] == np.datetime64("2021-01-01")

    dates = generate_normalized_sample_dates(
        np.datetime64("2020-01-01"), np.datetime64("2022-01-01"), Frequency.YEARLY
    )
    assert len(dates) == 3
    assert dates[0] == np.datetime64("2020-01-01")
    assert dates[-1] == np.datetime64("2022-01-01")

    dates = generate_normalized_sample_dates(
        np.datetime64("2020-12-30"),
        np.datetime64("2022-01-01T01:01"),
        Frequency.YEARLY,
    )
    assert len(dates) == 4
    assert dates[0] == np.datetime64("2020-01-01")
    assert dates[-1] == np.datetime64("2023-01-01")


def test_interpolate_backfill() -> None:

    raw_x = np.array([0, 2, 4, 6])
    raw_y = np.array([0, 20, 40, 60])

    x = np.array([0, 2, 4, 6])
    y = interpolate_backfill(x, raw_x, raw_y, -99, 99)
    assert (y == raw_y).all()

    x = np.array([-1, 1, 5, 7])
    expected_y = np.array([-99, 20, 60, 99])
    y = interpolate_backfill(x, raw_x, raw_y, -99, 99)
    assert (y == expected_y).all()

    x = np.array([-2, -1, 0, 3, 3, 6, 7, 8])
    expected_y = np.array([-99, -99, 0, 40, 40, 60, 99, 99])
    y = interpolate_backfill(x, raw_x, raw_y, -99, 99)
    assert (y == expected_y).all()


def test_sample_segmented_multi_real_table_at_date_with_single_real() -> None:
    # pylint: disable=too-many-statements
    # fmt:off
    input_data = [
        ["DATE",                             "REAL",  "T",   "R"],
        [np.datetime64("2020-01-01", "ms"),  1,       10.0,  1],
        [np.datetime64("2020-01-04", "ms"),  1,       40.0,  4],
        [np.datetime64("2020-01-06", "ms"),  1,       60.0,  6],
    ]
    # fmt:on

    schema = pa.schema(
        [
            pa.field("DATE", pa.timestamp("ms")),
            pa.field("REAL", pa.int64()),
            pa.field("T", pa.float32(), metadata={b"is_rate": b"False"}),
            pa.field("R", pa.float32(), metadata={b"is_rate": b"True"}),
        ]
    )

    table = _create_table_from_row_data(per_row_input_data=input_data, schema=schema)

    # Exact hit, first actual date
    sampledate = np.datetime64("2020-01-01", "ms")
    res = sample_segmented_multi_real_table_at_date(table, sampledate)
    assert res.num_rows == 1
    assert res["DATE"].to_numpy()[0] == sampledate
    assert res["REAL"][0].as_py() == 1
    assert res["T"][0].as_py() == 10
    assert res["R"][0].as_py() == 1

    # Exact hit, last actual date
    sampledate = np.datetime64("2020-01-06", "ms")
    res = sample_segmented_multi_real_table_at_date(table, sampledate)
    assert res.num_rows == 1
    assert res["DATE"].to_numpy()[0] == sampledate
    assert res["REAL"][0].as_py() == 1
    assert res["T"][0].as_py() == 60
    assert res["R"][0].as_py() == 6

    # Exact hit, middle date
    sampledate = np.datetime64("2020-01-04", "ms")
    res = sample_segmented_multi_real_table_at_date(table, sampledate)
    assert res.num_rows == 1
    assert res["DATE"].to_numpy()[0] == sampledate
    assert res["REAL"][0].as_py() == 1
    assert res["T"][0].as_py() == 40
    assert res["R"][0].as_py() == 4

    # Before first date
    sampledate = np.datetime64("2019-01-01", "ms")
    res = sample_segmented_multi_real_table_at_date(table, sampledate)
    assert res.num_rows == 1
    assert res["DATE"].to_numpy()[0] == sampledate
    assert res["REAL"][0].as_py() == 1
    assert res["T"][0].as_py() == 10
    assert res["R"][0].as_py() == 0

    # After last date
    sampledate = np.datetime64("2020-01-10", "ms")
    res = sample_segmented_multi_real_table_at_date(table, sampledate)
    assert res.num_rows == 1
    assert res["DATE"].to_numpy()[0] == sampledate
    assert res["REAL"][0].as_py() == 1
    assert res["T"][0].as_py() == 60
    assert res["R"][0].as_py() == 0

    # Interpolated
    sampledate = np.datetime64("2020-01-02", "ms")
    res = sample_segmented_multi_real_table_at_date(table, sampledate)
    assert res.num_rows == 1
    assert res["DATE"].to_numpy()[0] == sampledate
    assert res["REAL"][0].as_py() == 1
    assert res["T"][0].as_py() == 20
    assert res["R"][0].as_py() == 4

    # Interpolated
    sampledate = np.datetime64("2020-01-03", "ms")
    res = sample_segmented_multi_real_table_at_date(table, sampledate)
    assert res.num_rows == 1
    assert res["DATE"].to_numpy()[0] == sampledate
    assert res["REAL"][0].as_py() == 1
    assert res["T"][0].as_py() == 30
    assert res["R"][0].as_py() == 4


def test_sample_segmented_multi_real_table_at_date() -> None:
    # pylint: disable=too-many-statements
    # fmt:off
    input_data = [
        ["DATE",                             "REAL",  "T",    "R"],
        [np.datetime64("2020-01-01", "ms"),  0,       10.0,   1],
        [np.datetime64("2020-01-04", "ms"),  0,       40.0,   4],
        [np.datetime64("2020-01-06", "ms"),  0,       60.0,   6],
        [np.datetime64("2020-01-02", "ms"),  1,       2000.0,  200],
        [np.datetime64("2020-01-05", "ms"),  1,       5000.0,  500],
        [np.datetime64("2020-01-07", "ms"),  1,       7000.0,  700],
    ]
    # fmt:on

    schema = pa.schema(
        [
            pa.field("DATE", pa.timestamp("ms")),
            pa.field("REAL", pa.int64()),
            pa.field("T", pa.float32(), metadata={b"is_rate": b"False"}),
            pa.field("R", pa.float32(), metadata={b"is_rate": b"True"}),
        ]
    )

    table = _create_table_from_row_data(per_row_input_data=input_data, schema=schema)

    # Exact hit on first date in R=0
    sampledate = np.datetime64("2020-01-01", "ms")
    res = sample_segmented_multi_real_table_at_date(table, sampledate)
    assert res.num_rows == 2
    assert res["DATE"].to_numpy()[0] == sampledate
    assert res["DATE"].to_numpy()[1] == sampledate
    assert res["REAL"][0].as_py() == 0
    assert res["REAL"][1].as_py() == 1
    assert res["T"][0].as_py() == 10
    assert res["T"][1].as_py() == 2000
    assert res["R"][0].as_py() == 1
    assert res["R"][1].as_py() == 0

    # Exact hit on first date in R=1
    sampledate = np.datetime64("2020-01-02", "ms")
    res = sample_segmented_multi_real_table_at_date(table, sampledate)
    assert res.num_rows == 2
    assert res["DATE"].to_numpy()[0] == sampledate
    assert res["DATE"].to_numpy()[1] == sampledate
    assert res["REAL"][0].as_py() == 0
    assert res["REAL"][1].as_py() == 1
    assert res["T"][0].as_py() == 20
    assert res["T"][1].as_py() == 2000
    assert res["R"][0].as_py() == 4
    assert res["R"][1].as_py() == 200

    # Exact hit on last actual date in R=0
    sampledate = np.datetime64("2020-01-06", "ms")
    res = sample_segmented_multi_real_table_at_date(table, sampledate)
    assert res.num_rows == 2
    assert res["DATE"].to_numpy()[0] == sampledate
    assert res["DATE"].to_numpy()[1] == sampledate
    assert res["REAL"][0].as_py() == 0
    assert res["REAL"][1].as_py() == 1
    assert res["T"][0].as_py() == 60
    assert res["T"][1].as_py() == 6000
    assert res["R"][0].as_py() == 6
    assert res["R"][1].as_py() == 700

    # Exact hit on last actual date in R=1
    sampledate = np.datetime64("2020-01-07", "ms")
    res = sample_segmented_multi_real_table_at_date(table, sampledate)
    assert res.num_rows == 2
    assert res["DATE"].to_numpy()[0] == sampledate
    assert res["DATE"].to_numpy()[1] == sampledate
    assert res["REAL"][0].as_py() == 0
    assert res["REAL"][1].as_py() == 1
    assert res["T"][0].as_py() == 60
    assert res["T"][1].as_py() == 7000
    assert res["R"][0].as_py() == 0
    assert res["R"][1].as_py() == 700

    # Interpolated
    sampledate = np.datetime64("2020-01-02", "ms")
    res = sample_segmented_multi_real_table_at_date(table, sampledate)
    assert res.num_rows == 2
    assert res["DATE"].to_numpy()[0] == sampledate
    assert res["DATE"].to_numpy()[1] == sampledate
    assert res["REAL"][0].as_py() == 0
    assert res["REAL"][1].as_py() == 1
    assert res["T"][0].as_py() == 20
    assert res["T"][1].as_py() == 2000
    assert res["R"][0].as_py() == 4
    assert res["R"][1].as_py() == 200

    # Interpolated
    sampledate = np.datetime64("2020-01-03", "ms")
    res = sample_segmented_multi_real_table_at_date(table, sampledate)
    assert res.num_rows == 2
    assert res["DATE"].to_numpy()[0] == sampledate
    assert res["DATE"].to_numpy()[1] == sampledate
    assert res["REAL"][0].as_py() == 0
    assert res["REAL"][1].as_py() == 1
    assert res["T"][0].as_py() == 30
    assert res["T"][1].as_py() == 3000
    assert res["R"][0].as_py() == 4
    assert res["R"][1].as_py() == 500



from typing import Optional

import pyarrow as pa

from .ensemble_table_provider import ColumnMetadata


def create_column_metadata_from_field_meta(
    field: pa.Field,
) -> Optional[ColumnMetadata]:
    """Create VectorMetadata from keywords stored in the field's metadata"""

    meta_dict = field.metadata
    if not meta_dict:
        return None

    try:
        unit_bytestr = meta_dict[b"unit"]
    except KeyError:
        return ColumnMetadata(unit=None)

    return ColumnMetadata(
        unit=unit_bytestr.decode("ascii"),
    )



from typing import List

import webviz_core_components as wcc
from dash.development.base_component import Component
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import SettingsGroupABC


class Selectors(SettingsGroupABC):
    class IDs(StrEnum):
        RESPONSE = "response"

    def __init__(
        self,
        responses: List[str],
        initial_response: str,
    ) -> None:
        super().__init__("Selectors")
        self._responses = responses
        self._initial_response = initial_response

    def layout(self) -> List[Component]:
        return [
            wcc.Dropdown(
                id=self.register_component_unique_id(Selectors.IDs.RESPONSE),
                label="Response",
                options=[{"label": i, "value": i} for i in self._responses],
                value=self._initial_response,
                multi=False,
                clearable=False,
            )
        ]



import abc
from dataclasses import dataclass
from typing import List

import numpy as np
import xtgeo


@dataclass(frozen=True)
class WellPath:
    x_arr: np.ndarray
    y_arr: np.ndarray
    z_arr: np.ndarray
    md_arr: np.ndarray


# Class provides data for wells
class WellProvider(abc.ABC):
    @abc.abstractmethod
    def provider_id(self) -> str:
        """Returns string ID of the provider."""

    @abc.abstractmethod
    def well_names(self) -> List[str]:
        """Returns list of all available well names."""

    @abc.abstractmethod
    def get_well_path(self, well_name: str) -> WellPath:
        """Returns the coordinates for the well path along with MD for the well."""

    @abc.abstractmethod
    def get_well_xtgeo_obj(self, well_name: str) -> xtgeo.Well:
        ...



from typing import Dict, List, Optional

from webviz_subsurface._providers import (
    EnsembleFaultPolygonsProviderFactory,
    FaultPolygonsServer,
    SimulatedFaultPolygonsAddress,
)


class FaultPolygonsHandler:
    def __init__(
        self,
        server: FaultPolygonsServer,
        ensemble_path: str,
        map_surface_names_to_fault_polygons: Dict[str, str],
        fault_polygon_attribute: str,
    ) -> None:
        self._server = server
        polygon_provider_factory = EnsembleFaultPolygonsProviderFactory.instance()
        self._provider = (
            polygon_provider_factory.create_from_ensemble_fault_polygons_files(
                ensemble_path
            )
        )
        server.add_provider(self._provider)
        self._map_surface_names_to_fault_polygons = map_surface_names_to_fault_polygons
        self._fault_polygon_attribute = fault_polygon_attribute

    def extract_fault_polygon_url(
        self,
        polygon_name: str,
        realization: List[int],
    ) -> Optional[str]:
        if polygon_name is None:
            return None
        if len(realization) == 0:
            return None
        # NB! This always returns the url corresponding to the first realization
        address = SimulatedFaultPolygonsAddress(
            attribute=self._fault_polygon_attribute,
            name=self._map_surface_names_to_fault_polygons.get(
                polygon_name, polygon_name
            ),
            realization=realization[0],
        )
        return self._server.encode_partial_url(
            provider_id=self._provider.provider_id(),
            fault_polygons_address=address,
        )



from typing import Callable

import webviz_core_components as wcc
from dash import Dash, Input, Output, State, dash_table

from ..models import ParametersModel


def parameter_qc_controller(
    app: Dash, get_uuid: Callable, parametermodel: ParametersModel
):
    @app.callback(
        Output(get_uuid("property-qc-wrapper"), "children"),
        Input({"id": get_uuid("ensemble-selector"), "tab": "qc"}, "value"),
        Input({"id": get_uuid("delta-ensemble-selector"), "tab": "qc"}, "value"),
        Input(
            {"id": get_uuid("filter-parameter"), "tab": "qc"},
            "value",
        ),
        Input(get_uuid("property-qc-plot-type"), "value"),
    )
    def _update_bars(ensemble, delta_ensemble, parameters, plot_type):
        """Callback to switch visualization between table and distribution plots"""
        parameters = parameters if isinstance(parameters, list) else [parameters]
        ensembles = [ensemble, delta_ensemble]
        valid_params = parametermodel.pmodel.get_parameters_for_ensembles(ensembles)
        parameters = [x for x in parameters if x in valid_params]

        if plot_type == "table":
            columns, dframe = parametermodel.make_statistics_table(
                ensembles=ensembles, parameters=parameters
            )
            return dash_table.DataTable(
                style_table={
                    "height": "75vh",
                    "overflow": "auto",
                    "fontSize": 15,
                },
                style_cell={"textAlign": "center"},
                style_cell_conditional=[
                    {"if": {"column_id": "PARAMETER|"}, "textAlign": "left"}
                ],
                columns=columns,
                data=dframe,
                sort_action="native",
                filter_action="native",
                merge_duplicate_headers=True,
            )
        return wcc.Graph(
            id=get_uuid("property-qc-graph"),
            config={"displayModeBar": False},
            style={"height": "75vh"},
            figure=parametermodel.make_grouped_plot(
                ensembles=ensembles,
                parameters=parameters,
                plot_type=plot_type,
            ),
        )

    @app.callback(
        Output({"id": get_uuid("filter-parameter"), "tab": "qc"}, "options"),
        Output({"id": get_uuid("filter-parameter"), "tab": "qc"}, "value"),
        Input(get_uuid("delta-sort"), "value"),
        Input({"id": get_uuid("ensemble-selector"), "tab": "qc"}, "value"),
        Input({"id": get_uuid("delta-ensemble-selector"), "tab": "qc"}, "value"),
        State({"id": get_uuid("filter-parameter"), "tab": "qc"}, "value"),
    )
    def _update_parameters(sortby, ensemble, delta_ensemble, current_params):
        """Callback to sort parameters based on selection"""
        parametermodel.sort_parameters(
            ensemble=ensemble,
            delta_ensemble=delta_ensemble,
            sortby=sortby,
        )
        valid_params = parametermodel.pmodel.get_parameters_for_ensembles(
            [ensemble, delta_ensemble]
        )
        sorted_params = [x for x in parametermodel.parameters if x in valid_params]
        return (
            [{"label": i, "value": i} for i in sorted_params],
            [x for x in current_params if x in sorted_params],
        )



import webviz_core_components as wcc
from dash import dash_table, html


def uncertainty_table_layout(
    uuid: str,
) -> html.Div:
    """Layout for the uncertainty table dialog"""
    return html.Div(
        children=[
            wcc.FlexBox(
                children=[
                    html.Label(
                        className="webviz-structunc-uncertainty-table-label",
                        children="Statistics for well: ",
                        id={"id": uuid, "element": "label"},
                    ),
                    html.Button(
                        "Recalculate",
                        className="webviz-structunc-uncertainty-table-apply-btn",
                        id={"id": uuid, "element": "apply-button"},
                    ),
                ]
            ),
            dash_table.DataTable(
                id={"id": uuid, "element": "table"},
                columns=[
                    {"id": "Surface name", "name": "Surface name", "selectable": False},
                    {"id": "Ensemble", "name": "Ensemble", "selectable": False},
                    {"id": "Calculation", "name": "Calculation", "selectable": False},
                    {"id": "Pick no", "name": "Pick no", "selectable": False},
                    {
                        "id": "TVD",
                        "name": "TVD (MSL)",
                        "selectable": False,
                    },
                    {
                        "id": "MD",
                        "name": "MD (RKB)",
                        "selectable": False,
                    },
                ],
                style_header={
                    "opacity": 0.5,
                },
                style_filter={
                    "opacity": 0.5,
                },
                style_data_conditional=[
                    {
                        "if": {"column_id": "Surface name"},
                        "textAlign": "left",
                        "width": "15%",
                    },
                    {
                        "if": {"column_id": "Ensemble"},
                        "textAlign": "left",
                        "width": "15%",
                    },
                    {
                        "if": {"column_id": "Calculation"},
                        "textAlign": "left",
                        "width": "10%",
                    },
                    {
                        "if": {"filter_query": '{Calculation} = "Mean"'},
                        "backgroundColor": "rgba(0,177,106,0.3)",
                    },
                ],
                sort_action="native",
                filter_action="native",
            ),
        ],
    )



from typing import List, Optional

from webviz_subsurface_components import (
    ExpressionInfo,
    VariableVectorMapInfo,
    VectorDefinition,
)

from webviz_subsurface._providers import VectorMetadata
from webviz_subsurface._utils.ensemble_summary_provider_set import (
    EnsembleSummaryProviderSet,
)

# pylint: disable=line-too-long
from webviz_subsurface.plugins._simulation_time_series._views._subplot_view._utils.ensemble_summary_provider_set_utils import (
    create_calculated_unit_from_provider_set,
    create_vector_plot_titles_from_provider_set,
)

from ....mocks.ensemble_summary_provider_dummy import EnsembleSummaryProviderDummy

# *******************************************************************
#####################################################################
#
# CONFIGURE TESTDATA
#
#####################################################################
# *******************************************************************


class EnsembleSummaryProviderMock(EnsembleSummaryProviderDummy):
    """Ensemble summary provider mock for testing

    Note empty list returned in override methods, only to allow constructing
    EnsembleSummaryProviderSet objects!
    """

    ########################################
    #
    # Override methods
    #
    ########################################
    def vector_names(self) -> List[str]:
        """Return empty list only to allow constructing EnsembleSummaryProviderSet object"""
        return ["FGIT", "WGOR:A1", "WBHP:A1", "WOPT:A1"]

    def realizations(self) -> List[int]:
        """Return empty list only to allow constructing EnsembleSummaryProviderSet object"""
        return []

    def vector_metadata(self, vector_name: str) -> Optional[VectorMetadata]:
        # NOTE: All vector names below must be defined in vector_names() method as well!
        if vector_name == "FGIT":
            return VectorMetadata(
                unit="unit_1",
                is_total=False,
                is_rate=False,
                is_historical=False,
                keyword="FGIT",
                wgname=None,
                get_num=None,
            )
        if vector_name == "WGOR:A1":
            return VectorMetadata(
                unit="unit_2",
                is_total=False,
                is_rate=False,
                is_historical=False,
                keyword="A1",
                wgname=None,
                get_num=None,
            )
        if vector_name == "WBHP:A1":
            return VectorMetadata(
                unit="unit_3",
                is_total=False,
                is_rate=False,
                is_historical=False,
                keyword="A1",
                wgname=None,
                get_num=None,
            )
        if vector_name == "WOPT:A1":
            return VectorMetadata(
                unit="unit_4",
                is_total=False,
                is_rate=False,
                is_historical=False,
                keyword="A1",
                wgname=None,
                get_num=None,
            )
        return None


FIRST_TEST_EXPRESSION = ExpressionInfo(
    name="First Expression",
    expression="x+y",
    id="FirstId",
    variableVectorMap=[
        VariableVectorMapInfo(variableName="x", vectorName=["FGIT"]),
        VariableVectorMapInfo(variableName="y", vectorName=["WGOR:A1"]),
    ],
    isValid=True,
    isDeletable=False,
)

SECOND_TEST_EXPRESSION = ExpressionInfo(
    name="Second Expression",
    expression="x-y",
    id="SecondId",
    variableVectorMap=[
        VariableVectorMapInfo(variableName="x", vectorName=["FGIT"]),
        VariableVectorMapInfo(variableName="y", vectorName=["WBHP:A1"]),
    ],
    isValid=True,
    isDeletable=False,
)

THIRD_TEST_EXPRESSION = ExpressionInfo(
    name="Third Expression",
    expression="x/y",
    id="ThirdId",
    variableVectorMap=[
        VariableVectorMapInfo(variableName="x", vectorName=["WBHP:A1"]),
        VariableVectorMapInfo(variableName="y", vectorName=["WGOR:A1"]),
    ],
    isValid=True,
    isDeletable=False,
)

INVALID_TEST_EXPRESSION = ExpressionInfo(
    name="Invalid Expression",
    expression="invalid_function(x)",  # Invalid function
    id="InvalidId",
    variableVectorMap=[
        VariableVectorMapInfo(variableName="x", vectorName=["WBHP:A1"]),
    ],
    isValid=True,
    isDeletable=False,
)

TEST_PROVIDER_SET = EnsembleSummaryProviderSet(
    {"First Provider": EnsembleSummaryProviderMock()}
)


# *******************************************************************
#####################################################################
#
# UNIT TESTS
#
#####################################################################
# *******************************************************************


def test_create_calculated_unit_from_provider_set() -> None:
    assert (
        create_calculated_unit_from_provider_set(
            FIRST_TEST_EXPRESSION, TEST_PROVIDER_SET
        )
        == "unit_1+unit_2"
    )
    assert (
        create_calculated_unit_from_provider_set(
            SECOND_TEST_EXPRESSION, TEST_PROVIDER_SET
        )
        == "unit_1-unit_3"
    )
    assert (
        create_calculated_unit_from_provider_set(
            THIRD_TEST_EXPRESSION, TEST_PROVIDER_SET
        )
        == "unit_3/unit_2"
    )
    assert (
        create_calculated_unit_from_provider_set(
            INVALID_TEST_EXPRESSION, TEST_PROVIDER_SET
        )
        is None
    )


def test_create_vector_plot_titles_from_provider_set() -> None:
    vector_names = ["FGIT", "WGOR:A1", "PER_DAY_WOPT:A1", "First Expression"]
    expressions = [FIRST_TEST_EXPRESSION]

    # Test WITHOUT user defined vector definitions
    expected_titles = {
        "FGIT": "Gas Injection Total [unit_1]",
        "WGOR:A1": "Gas-Oil Ratio, well A1 [unit_2]",
        "PER_DAY_WOPT:A1": "Average Oil Production Total Per day, well A1 [unit_4/DAY]",
        "First Expression": "First Expression [unit_1+unit_2]",
    }
    assert expected_titles == create_vector_plot_titles_from_provider_set(
        vector_names, expressions, TEST_PROVIDER_SET, {}
    )

    # Test WITH user defined vector definitions
    user_defined_definitions = {
        "FGIT": VectorDefinition(description="First Test title", type="field"),
        "WOPT": VectorDefinition(description="Second Test title", type="well"),
    }
    expected_titles = {
        "FGIT": "First Test title [unit_1]",
        "WGOR:A1": "Gas-Oil Ratio, well A1 [unit_2]",
        "PER_DAY_WOPT:A1": "Average Second Test title Per day, well A1 [unit_4/DAY]",
        "First Expression": "First Expression [unit_1+unit_2]",
    }

    assert expected_titles == create_vector_plot_titles_from_provider_set(
        vector_names, expressions, TEST_PROVIDER_SET, user_defined_definitions
    )



import glob
import os
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional

# The fmu.ensemble dependency ecl is only available for Linux,
# hence, ignore any import exception here to make
# it still possible to use the PvtPlugin on
# machines with other OSes.
#
# NOTE: Functions in this file cannot be used
#       on non-Linux OSes.
try:
    from fmu.ensemble import ScratchEnsemble
except ImportError:
    pass


@dataclass(frozen=True)
class SurfaceFileInfo:
    path: str
    real: int
    name: str
    attribute: str
    datestr: Optional[str]


def _discover_ensemble_realizations_fmu(ens_path: str) -> Dict[int, str]:
    """Returns dict indexed by realization number and with runpath as value"""
    scratch_ensemble = ScratchEnsemble("dummyEnsembleName", paths=ens_path).filter("OK")
    real_dict = {i: r.runpath() for i, r in scratch_ensemble.realizations.items()}
    return real_dict


def _discover_ensemble_realizations(ens_path: str) -> Dict[int, str]:
    # Much faster than FMU impl above, but is it risky?
    # Do we need to check for OK-file?
    real_dict: Dict[int, str] = {}

    realidxregexp = re.compile(r"realization-(\d+)")
    globbed_real_dirs = sorted(glob.glob(str(ens_path)))
    for real_dir in globbed_real_dirs:
        realnum: Optional[int] = None
        for path_comp in reversed(real_dir.split(os.path.sep)):
            realmatch = re.match(realidxregexp, path_comp)
            if realmatch:
                realnum = int(realmatch.group(1))
                break

        if realnum is not None:
            real_dict[realnum] = real_dir

    return real_dict


@dataclass(frozen=True)
class SurfaceIdent:
    name: str
    attribute: str
    datestr: Optional[str]


def _surface_ident_from_filename(filename: str) -> Optional[SurfaceIdent]:
    """Split the stem part of the surface filename into surface name, attribute and
    optionally date part"""
    delimiter: str = "--"
    parts = Path(filename).stem.split(delimiter)
    if len(parts) < 2:
        return None

    return SurfaceIdent(
        name=parts[0], attribute=parts[1], datestr=parts[2] if len(parts) >= 3 else None
    )


def discover_per_realization_surface_files(
    ens_path: str,
    rel_surface_folder: str,
    attribute_filter: List[str] = None,
) -> List[SurfaceFileInfo]:
    suffix: str = "*.gri"

    surface_files: List[SurfaceFileInfo] = []

    real_dict = _discover_ensemble_realizations_fmu(ens_path)
    for realnum, runpath in sorted(real_dict.items()):
        globbed_filenames = glob.glob(str(Path(runpath) / rel_surface_folder / suffix))
        for surf_filename in sorted(globbed_filenames):
            surf_ident = _surface_ident_from_filename(surf_filename)
            if surf_ident:
                if (
                    attribute_filter is not None
                    and surf_ident.attribute not in attribute_filter
                ):
                    continue
                surface_files.append(
                    SurfaceFileInfo(
                        path=surf_filename,
                        real=realnum,
                        name=surf_ident.name,
                        attribute=surf_ident.attribute,
                        datestr=surf_ident.datestr,
                    )
                )

    return surface_files


def discover_observed_surface_files(
    ens_path: str, attribute_filter: List[str] = None
) -> List[SurfaceFileInfo]:
    observed_surface_folder: str = "share/observations/maps"
    suffix: str = "*.gri"

    surface_files: List[SurfaceFileInfo] = []

    ens_root_path = ens_path.split("realization")[0]
    globbed_filenames = glob.glob(
        str(Path(ens_root_path) / observed_surface_folder / suffix)
    )
    for surf_filename in sorted(globbed_filenames):
        surf_ident = _surface_ident_from_filename(surf_filename)
        if surf_ident:
            if (
                attribute_filter is not None
                and surf_ident.attribute not in attribute_filter
            ):
                continue
            surface_files.append(
                SurfaceFileInfo(
                    path=surf_filename,
                    real=-1,
                    name=surf_ident.name,
                    attribute=surf_ident.attribute,
                    datestr=surf_ident.datestr,
                )
            )

    return surface_files



import webviz_core_components as wcc
from webviz_config.utils import StrEnum
from webviz_config.webviz_plugin_subclasses import ViewElementABC


class TornadoPlot(ViewElementABC):
    class IDs(StrEnum):
        GRAPH = "graph"

    def __init__(self, height: str = "86vh") -> None:
        super().__init__()
        self.height = height

    def inner_layout(self) -> wcc.Graph:
        return wcc.Graph(
            id=self.register_component_unique_id(TornadoPlot.IDs.GRAPH),
            config={"displayModeBar": True},
            style={"display": "block", "height": self.height},
        )



import datetime
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Tuple

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.compute as pc

from webviz_subsurface._utils.perf_timer import PerfTimer

from ._field_metadata import create_vector_metadata_from_field_meta
from ._resampling import (
    generate_normalized_sample_dates,
    resample_segmented_multi_real_table,
    sample_segmented_multi_real_table_at_date,
)
from ._table_utils import (
    add_per_vector_min_max_to_table_schema_metadata,
    find_intersected_dates_between_realizations,
    find_min_max_for_numeric_table_columns,
    get_per_vector_min_max_from_schema_metadata,
)
from .ensemble_summary_provider import (
    EnsembleSummaryProvider,
    Frequency,
    VectorMetadata,
)

# Since PyArrow's actual compute functions are not seen by pylint
# pylint: disable=no-member

LOGGER = logging.getLogger(__name__)


def _sort_table_on_real_then_date(table: pa.Table) -> pa.Table:
    indices = pc.sort_indices(
        table, sort_keys=[("REAL", "ascending"), ("DATE", "ascending")]
    )
    sorted_table = table.take(indices)
    return sorted_table


def _is_date_column_monotonically_increasing(table: pa.Table) -> bool:
    dates_np = table.column("DATE").to_numpy()
    if not np.all(np.diff(dates_np) > np.timedelta64(0)):
        return False

    return True


def _find_first_non_increasing_date_pair(
    table: pa.Table,
) -> Tuple[Optional[np.datetime64], Optional[np.datetime64]]:
    dates_np = table.column("DATE").to_numpy()
    offending_indices = np.asarray(np.diff(dates_np) <= np.timedelta64(0)).nonzero()[0]
    if not offending_indices:
        return (None, None)

    return (dates_np[offending_indices[0]], dates_np[offending_indices[0] + 1])


class ProviderImplArrowLazy(EnsembleSummaryProvider):
    """This class implements an EnsembleSummaryProvider with lazy (on-demand)
    resampling/interpolation.
    """

    def __init__(self, arrow_file_name: Path) -> None:
        self._arrow_file_name = str(arrow_file_name)

        LOGGER.debug(f"init with arrow file: {self._arrow_file_name}")
        timer = PerfTimer()

        source = pa.memory_map(self._arrow_file_name, "r")
        et_open_ms = timer.lap_ms()

        reader = pa.ipc.RecordBatchFileReader(source)
        et_create_reader_ms = timer.lap_ms()

        # Discover columns and realizations that are present in the file
        column_names_on_file = reader.schema.names
        self._vector_names: List[str] = [
            colname
            for colname in column_names_on_file
            if colname not in ["DATE", "REAL", "ENSEMBLE"]
        ]
        et_find_vec_names_ms = timer.lap_ms()

        unique_realizations_on_file = reader.read_all().column("REAL").unique()
        self._realizations: List[int] = unique_realizations_on_file.to_pylist()
        et_find_real_ms = timer.lap_ms()

        # We'll try and keep the file open for the life-span of the provider.
        # Done to try and stop blobfuse from throwing the file out of its cache.
        self._cached_reader = reader

        # For testing, uncomment code below and we will be more aggressive
        # and keep the "raw" table in memory
        self._cached_full_table = None
        # self._cached_full_table = reader.read_all()

        LOGGER.debug(
            f"init took: {timer.elapsed_s():.2f}s, "
            f"(open={et_open_ms}ms, create_reader={et_create_reader_ms}ms, "
            f"find_vec_names={et_find_vec_names_ms}ms, find_real={et_find_real_ms}ms), "
            f"#vector_names={len(self._vector_names)}, "
            f"#realization={len(self._realizations)}"
        )

        if not self._realizations:
            raise ValueError("Init from backing store failed NO realizations")
        if not self._vector_names:
            raise ValueError("Init from backing store failed NO vector_names")

    @staticmethod
    def write_backing_store_from_per_realization_tables(
        storage_dir: Path, storage_key: str, per_real_tables: Dict[int, pa.Table]
    ) -> None:
        # pylint: disable=too-many-locals
        @dataclass
        class Elapsed:
            concat_tables_s: float = -1
            build_add_real_col_s: float = -1
            sorting_s: float = -1
            find_and_store_min_max_s: float = -1
            write_s: float = -1

        elapsed = Elapsed()

        arrow_file_name = storage_dir / (storage_key + ".arrow")
        LOGGER.debug(f"Writing backing store to arrow file: {arrow_file_name}")
        timer = PerfTimer()

        unique_column_names = set()
        for real_num, table in per_real_tables.items():
            unique_column_names.update(table.schema.names)

            if "REAL" in table.schema.names:
                raise ValueError(
                    f"Input tables should not have REAL column (real={real_num})"
                )

            if table.schema.field("DATE").type != pa.timestamp("ms"):
                raise ValueError(
                    f"DATE column must have timestamp[ms] data type (real={real_num})"
                )

            if not _is_date_column_monotonically_increasing(table):
                offending_pair = _find_first_non_increasing_date_pair(table)
                raise ValueError(
                    f"DATE column must be monotonically increasing\n"
                    f"Error detected in realization: {real_num}\n"
                    f"First offending timestamps: {offending_pair}"
                )

        LOGGER.debug(
            f"Concatenating {len(per_real_tables)} tables with "
            f"{len(unique_column_names)} unique column names"
        )

        full_table = pa.concat_tables(per_real_tables.values(), promote=True)
        elapsed.concat_tables_s = timer.lap_s()

        real_arr = np.empty(full_table.num_rows, np.int32)
        table_start_idx = 0
        for real_num, real_table in per_real_tables.items():
            real_arr[table_start_idx : table_start_idx + real_table.num_rows] = real_num
            table_start_idx += real_table.num_rows

        full_table = full_table.add_column(0, "REAL", pa.array(real_arr))
        elapsed.build_add_real_col_s = timer.lap_s()

        # Must sort table on real since interpolations work per realization
        # and we utilize slicing for speed
        full_table = _sort_table_on_real_then_date(full_table)
        elapsed.sorting_s = timer.lap_s()

        # Find per column min/max values and store them as metadata on table's schema
        per_vector_min_max = find_min_max_for_numeric_table_columns(full_table)
        full_table = add_per_vector_min_max_to_table_schema_metadata(
            full_table, per_vector_min_max
        )
        elapsed.find_and_store_min_max_s = timer.lap_s()

        # feather.write_feather(full_table, dest=arrow_file_name)
        with pa.OSFile(str(arrow_file_name), "wb") as sink:
            with pa.RecordBatchFileWriter(sink, full_table.schema) as writer:
                writer.write_table(full_table)
        elapsed.write_s = timer.lap_s()

        LOGGER.debug(
            f"Wrote backing store to arrow file in: {timer.elapsed_s():.2f}s ("
            f"concat_tables={elapsed.concat_tables_s:.2f}s, "
            f"build_add_real_col={elapsed.build_add_real_col_s:.2f}s, "
            f"sorting={elapsed.sorting_s:.2f}s, "
            f"find_and_store_min_max={elapsed.find_and_store_min_max_s:.2f}s, "
            f"write={elapsed.write_s:.2f}s)"
        )

    @staticmethod
    def from_backing_store(
        storage_dir: Path, storage_key: str
    ) -> Optional["ProviderImplArrowLazy"]:

        arrow_file_name = storage_dir / (storage_key + ".arrow")
        if arrow_file_name.is_file():
            return ProviderImplArrowLazy(arrow_file_name)

        return None

    def _get_or_read_schema(self) -> pa.Schema:
        if self._cached_full_table:
            return self._cached_full_table.schema
        if self._cached_reader:
            return self._cached_reader.schema

        source = pa.memory_map(self._arrow_file_name, "r")
        return pa.ipc.RecordBatchFileReader(source).schema

    def _get_or_read_table(self, columns: List[str]) -> pa.Table:
        if self._cached_full_table:
            return self._cached_full_table.select(columns)
        if self._cached_reader:
            return self._cached_reader.read_all().select(columns)

        source = pa.memory_map(self._arrow_file_name, "r")
        reader = pa.ipc.RecordBatchFileReader(source)
        return reader.read_all().select(columns)

    def vector_names(self) -> List[str]:
        return self._vector_names

    def vector_names_filtered_by_value(
        self,
        exclude_all_values_zero: bool = False,
        exclude_constant_values: bool = False,
    ) -> List[str]:

        timer = PerfTimer()

        schema = self._get_or_read_schema()
        et_read_ms = timer.lap_ms()

        per_vector_min_max = get_per_vector_min_max_from_schema_metadata(schema)
        et_get_min_max_ms = timer.lap_ms()

        ret_vec_names: List[str] = []
        for vec_name in self._vector_names:
            minval = per_vector_min_max[vec_name]["min"]
            maxval = per_vector_min_max[vec_name]["max"]

            if minval == maxval:
                if exclude_constant_values:
                    continue

                if exclude_all_values_zero and minval == 0:
                    continue

            ret_vec_names.append(vec_name)
        et_filter_ms = timer.lap_ms()

        LOGGER.debug(
            f"vector_names_filtered_by_value() took: {timer.elapsed_ms()}ms ("
            f"read={et_read_ms}ms, "
            f"get_min_max={et_get_min_max_ms}ms, "
            f"filter={et_filter_ms}ms)"
        )

        return ret_vec_names

    def realizations(self) -> List[int]:
        return self._realizations

    def vector_metadata(self, vector_name: str) -> Optional[VectorMetadata]:
        schema = self._get_or_read_schema()
        field = schema.field(vector_name)
        return create_vector_metadata_from_field_meta(field)

    def supports_resampling(self) -> bool:
        return True

    def dates(
        self,
        resampling_frequency: Optional[Frequency],
        realizations: Optional[Sequence[int]] = None,
    ) -> List[datetime.datetime]:

        timer = PerfTimer()

        table = self._get_or_read_table(["DATE", "REAL"])
        et_read_ms = timer.lap_ms()

        if realizations:
            mask = pc.is_in(table["REAL"], value_set=pa.array(realizations))
            table = table.filter(mask)
        et_filter_ms = timer.lap_ms()

        if resampling_frequency is not None:
            unique_dates_np = table.column("DATE").unique().to_numpy()
            min_raw_date = np.min(unique_dates_np)
            max_raw_date = np.max(unique_dates_np)
            intersected_dates = generate_normalized_sample_dates(
                min_raw_date, max_raw_date, resampling_frequency
            )
        else:
            intersected_dates = find_intersected_dates_between_realizations(table)

        et_find_unique_ms = timer.lap_ms()

        LOGGER.debug(
            f"dates({resampling_frequency}) took: {timer.elapsed_ms()}ms ("
            f"read={et_read_ms}ms, "
            f"filter={et_filter_ms}ms, "
            f"find_unique={et_find_unique_ms}ms)"
        )

        return intersected_dates.astype(datetime.datetime).tolist()

    def get_vectors_df(
        self,
        vector_names: Sequence[str],
        resampling_frequency: Optional[Frequency],
        realizations: Optional[Sequence[int]] = None,
    ) -> pd.DataFrame:

        if not vector_names:
            raise ValueError("List of requested vector names is empty")

        timer = PerfTimer()

        columns_to_get = ["DATE", "REAL"]
        columns_to_get.extend(vector_names)
        table = self._get_or_read_table(columns_to_get)
        et_read_ms = timer.lap_ms()

        if realizations is not None:
            mask = pc.is_in(table["REAL"], value_set=pa.array(realizations))
            table = table.filter(mask)
        et_filter_ms = timer.lap_ms()

        if resampling_frequency is not None:
            table = resample_segmented_multi_real_table(table, resampling_frequency)
        et_resample_ms = timer.lap_ms()

        df = table.to_pandas(timestamp_as_object=True)
        et_to_pandas_ms = timer.lap_ms()

        LOGGER.debug(
            f"get_vectors_df({resampling_frequency}) took: {timer.elapsed_ms()}ms ("
            f"read={et_read_ms}ms, "
            f"filter={et_filter_ms}ms, "
            f"resample={et_resample_ms}ms, "
            f"to_pandas={et_to_pandas_ms}ms), "
            f"#vecs={len(vector_names)}, "
            f"#real={len(realizations) if realizations is not None else 'all'}, "
            f"df.shape={df.shape}, file={Path(self._arrow_file_name).name}"
        )

        return df

    def get_vectors_for_date_df(
        self,
        date: datetime.datetime,
        vector_names: Sequence[str],
        realizations: Optional[Sequence[int]] = None,
    ) -> pd.DataFrame:

        if not vector_names:
            raise ValueError("List of requested vector names is empty")

        timer = PerfTimer()

        columns_to_get = ["DATE", "REAL"]
        columns_to_get.extend(vector_names)
        table = self._get_or_read_table(columns_to_get)
        et_read_ms = timer.lap_ms()

        if realizations:
            real_mask = pc.is_in(table["REAL"], value_set=pa.array(realizations))
            table = table.filter(real_mask)
        et_filter_ms = timer.lap_ms()

        np_lookup_date = np.datetime64(date, "ms")
        table = sample_segmented_multi_real_table_at_date(table, np_lookup_date)

        et_resample_ms = timer.lap_ms()
        table = table.drop(["DATE"])

        df = table.to_pandas()
        et_to_pandas_ms = timer.lap_ms()

        LOGGER.debug(
            f"get_vectors_for_date_df() took: {timer.elapsed_ms()}ms ("
            f"read={et_read_ms}ms, "
            f"filter={et_filter_ms}ms, "
            f"resample={et_resample_ms}ms, "
            f"to_pandas={et_to_pandas_ms}ms), "
            f"#vecs={len(vector_names)}, "
            f"#real={len(realizations) if realizations else 'all'}, "
            f"df.shape={df.shape}, file={Path(self._arrow_file_name).name}"
        )

        return df



import io

import numpy as np
import xtgeo


def surface_to_float32_array(surface: xtgeo.RegularSurface) -> io.BytesIO:
    values = surface.values.astype(np.float32)
    values.fill_value = np.NaN
    values = np.ma.filled(values)

    # Rotate 90 deg left.
    # This will cause the width of to run along the X axis
    # and height of along Y axis (starting from bottom.)
    values = np.rot90(values)

    byte_io = io.BytesIO()
    byte_io.write(values.tobytes())
    byte_io.seek(0)
    return byte_io



from pathlib import Path
from typing import Callable, Dict, List, Tuple

import pandas as pd
import webviz_core_components as wcc
from dash import Dash, html
from webviz_config import WebvizPluginABC, WebvizSettings
from webviz_config.webviz_assets import WEBVIZ_ASSETS
from webviz_config.webviz_store import webvizstore

import webviz_subsurface
from webviz_subsurface._components.parameter_filter import ParameterFilter
from webviz_subsurface._models import ObservationModel
from webviz_subsurface._providers import EnsembleTableProviderFactory
from webviz_subsurface._utils.unique_theming import unique_colors

from .controllers import build_figure, update_figure_clientside
from .views import main_view


class LinePlotterFMU(WebvizPluginABC):
    """General line plotter for FMU data

    ---

    * **`ensembles`:** Which ensembles in `shared_settings` to visualize.
    * **`csvfile`:** Relative path to Csv file stored per realization
     * **`observation_file`:** Yaml file with observations
    * **`observation_group`:** Top-level key in observation file.
    * **`remap_observation_keys`:** Remap observation keys to columns in csv file
    * **`remap_observation_values`:** Remap observation values to columns in csv file
    * **`colors`:** Set colors for each ensemble
    * **`initial_data`:** Initialize data selectors (x,y,ensemble, parameter)
    * **`initial_layout`:** Initialize plot layout (x and y axis direction and type)"""

    # pylint: disable=too-many-locals, too-many-arguments
    def __init__(
        self,
        app: Dash,
        webviz_settings: WebvizSettings,
        csvfile: str = None,
        ensembles: list = None,
        aggregated_csvfile: Path = None,
        aggregated_parameterfile: Path = None,
        observation_file: Path = None,
        observation_group: str = "general",
        remap_observation_keys: Dict[str, str] = None,
        remap_observation_values: Dict[str, str] = None,
        colors: Dict = None,
        initial_data: Dict = None,
        initial_layout: Dict = None,
    ):
        super().__init__()

        provider_factory = EnsembleTableProviderFactory.instance()
        self._initial_data = initial_data if initial_data else {}
        self._initial_layout = initial_layout if initial_layout else {}
        if ensembles is not None and csvfile is not None:
            ensembles_dict: Dict[str, str] = {
                ens_name: webviz_settings.shared_settings["scratch_ensembles"][ens_name]
                for ens_name in ensembles
            }
            self._parameterproviderset = {
                ens_name: provider_factory.create_from_per_realization_parameter_file(
                    ens_path
                )
                for ens_name, ens_path in ensembles_dict.items()
            }
            self._tableproviderset = {
                ens_name: provider_factory.create_from_per_realization_csv_file(
                    ens_path, csvfile
                )
                for ens_name, ens_path in ensembles_dict.items()
            }
            self._ensemble_names = ensembles
        elif aggregated_csvfile and aggregated_parameterfile is not None:
            self._tableproviderset = (
                provider_factory.create_provider_set_from_aggregated_csv_file(
                    aggregated_csvfile
                )
            )
            self._parameterproviderset = (
                provider_factory.create_provider_set_from_aggregated_csv_file(
                    aggregated_parameterfile
                )
            )
            self._ensemble_names = list(self._tableproviderset.keys())
        else:
            raise ValueError(
                "Specify either ensembles and csvfile or aggregated_csvfile "
                "and aggregated_parameterfile"
            )
        all_parameters: list = [
            provider.column_names() for provider in self._parameterproviderset.values()
        ]
        self._parameter_names: list = list(set().union(*all_parameters))
        all_data_columns: list = [
            provider.column_names() for provider in self._tableproviderset.values()
        ]
        self._data_column_names: list = list(set().union(*all_data_columns))

        dfs = []
        for ens, provider in self._parameterproviderset.items():
            df = provider.get_column_data(column_names=provider.column_names())
            df["ENSEMBLE"] = ens
            dfs.append(df)
        parameterdf = pd.concat(dfs)

        self._realizations = sorted(list(parameterdf["REAL"].unique()))
        self._parameter_filter = ParameterFilter(
            self.uuid("parameter-filter"), parameterdf
        )
        self._observationfile = observation_file
        self._observationmodel = (
            ObservationModel(
                get_path(self._observationfile),
                observation_group,
                remap_observation_keys,
                remap_observation_values,
            )
            if self._observationfile
            else None
        )
        WEBVIZ_ASSETS.add(
            Path(webviz_subsurface.__file__).parent
            / "_assets"
            / "js"
            / "clientside_functions.js"
        )

        self._colors: Dict = unique_colors(self._ensemble_names, webviz_settings.theme)
        if colors is not None:
            self._colors.update(colors)

        self.set_callbacks(app)

    def add_webvizstore(self) -> List[Tuple[Callable, list]]:
        return (
            [(get_path, [{"path": self._observationfile}])]
            if self._observationfile is not None
            else []
        )

    @property
    def layout(self) -> html.Div:
        return wcc.FlexBox(
            children=[
                html.Div(
                    style={"flex": 5},
                    children=main_view(
                        get_uuid=self.uuid,
                        ensemble_names=self._ensemble_names,
                        data_column_names=self._data_column_names,
                        parameter_names=self._parameter_names,
                        realizations=self._realizations,
                        initial_data=self._initial_data,
                        initial_layout=self._initial_layout,
                    ),
                ),
                wcc.Frame(
                    style={"flex": 1, "height": "90vh"},
                    children=self._parameter_filter.layout,
                ),
            ]
        )

    def set_callbacks(self, app: Dash) -> None:
        build_figure(
            app,
            get_uuid=self.uuid,
            tableproviders=self._tableproviderset,
            observationmodel=self._observationmodel,
            parameterproviders=self._parameterproviderset,
            colors=self._colors,
        )
        update_figure_clientside(app, get_uuid=self.uuid)


@webvizstore
def get_path(path: Path) -> Path:
    return Path(path)



from pathlib import Path

import numpy as np
import pytest
import xtgeo

from webviz_subsurface._models import WellSetModel


@pytest.mark.usefixtures("app")
def test_well_set_model(testdata_folder: Path) -> None:
    wellfiles = [
        testdata_folder / "reek_test_data" / "observed_data" / "wells" / well
        for well in ["OP_1.w", "OP_2.w", "OP_3.w", "OP_4.w", "OP_5.w", "OP_6.w"]
    ]

    wmodel = WellSetModel(wellfiles=wellfiles)
    assert set(wmodel.well_names) == set(
        ["OP_1", "OP_2", "OP_3", "OP_4", "OP_5", "OP_6"]
    )
    for name, well in wmodel.wells.items():
        assert isinstance(name, str)
        assert isinstance(well, xtgeo.Well)
    op_6 = wmodel.get_well("OP_6")
    assert isinstance(op_6, xtgeo.Well)
    assert op_6.name == "OP_6"


@pytest.mark.usefixtures("app")
def test_logs(testdata_folder: Path) -> None:
    wmodel = WellSetModel(
        wellfiles=[
            testdata_folder / "reek_test_data" / "observed_data" / "wells" / "OP_6.w"
        ],
        zonelog="Zonelog",
    )
    well = wmodel.get_well("OP_6")
    assert well.zonelogname == "Zonelog"


@pytest.mark.usefixtures("app")
def test_tvd_truncation(testdata_folder: Path) -> None:
    wmodel = WellSetModel(
        wellfiles=[
            testdata_folder / "reek_test_data" / "observed_data" / "wells" / "OP_6.w"
        ],
        tvdmin=1000,
        tvdmax=1500,
    )
    well = wmodel.get_well("OP_6")
    assert well.dataframe["Z_TVDSS"].min() >= 1000
    assert well.dataframe["Z_TVDSS"].max() <= 1501


@pytest.mark.usefixtures("app")
def test_get_fence(testdata_folder: Path) -> None:
    wmodel = WellSetModel(
        wellfiles=[
            testdata_folder / "reek_test_data" / "observed_data" / "wells" / "OP_6.w"
        ],
        zonelog="Zonelog",
    )
    fence = wmodel.get_fence("OP_6")
    assert isinstance(fence, np.ndarray)
    # Test horizontal length
    assert int(fence[:, 3].min()) == -40
    assert int(fence[:, 3].max()) == 2713
    # Test tvd
    assert int(fence[:, 2].min()) == 1
    assert int(fence[:, 2].max()) == 1643



import abc
from dataclasses import dataclass
from typing import List, Optional

import xtgeo


@dataclass(frozen=True)
class SimulatedFaultPolygonsAddress:
    """Specifies a unique simulated fault polygon set for a given ensemble realization"""

    attribute: str
    name: str
    realization: int


# Type aliases used for signature readability
FaultPolygonsAddress = SimulatedFaultPolygonsAddress


# Class provides data for ensemble surfaces
class EnsembleFaultPolygonsProvider(abc.ABC):
    @abc.abstractmethod
    def provider_id(self) -> str:
        """Returns string ID of the provider."""

    @abc.abstractmethod
    def attributes(self) -> List[str]:
        """Returns list of all available attributes."""

    @abc.abstractmethod
    def fault_polygons_names_for_attribute(
        self, fault_polygons_attribute: str
    ) -> List[str]:
        """Returns list of all available fault polygons names for a given attribute."""

    @abc.abstractmethod
    def realizations(self) -> List[int]:
        """Returns list of all available realizations."""

    @abc.abstractmethod
    def get_fault_polygons(
        self,
        address: FaultPolygonsAddress,
    ) -> Optional[xtgeo.Polygons]:
        """Returns fault polygons for a given fault polygons address"""

    # @abc.abstractmethod
    # def get_surface_bounds(self, surface: EnsembleSurfaceContext) -> List[float]:
    #     """Returns the bounds for a surface [xmin,ymin, xmax,ymax]"""

    # @abc.abstractmethod
    # def get_surface_value_range(self, surface: EnsembleSurfaceContext) -> List[float]:
    #     """Returns the value range for a given surface context [zmin, zmax]"""

    # @abc.abstractmethod
    # def get_surface_as_rgba(self, surface: EnsembleSurfaceContext) -> io.BytesIO:
    #     """Returns surface as a greyscale png RGBA with encoded elevation values
    #     in a bytestream"""



import logging
import shutil
import warnings
from enum import Enum
from pathlib import Path
from typing import List, Optional, Set

import numpy as np
import pandas as pd
import xtgeo

from webviz_subsurface._utils.perf_timer import PerfTimer

from ._stat_surf_cache import StatSurfCache
from ._surface_discovery import SurfaceFileInfo
from .ensemble_surface_provider import (
    EnsembleSurfaceProvider,
    ObservedSurfaceAddress,
    SimulatedSurfaceAddress,
    StatisticalSurfaceAddress,
    SurfaceAddress,
    SurfaceStatistic,
)

LOGGER = logging.getLogger(__name__)

REL_SIM_DIR = "sim"
REL_OBS_DIR = "obs"
REL_STAT_CACHE_DIR = "stat_cache"

# pylint: disable=too-few-public-methods
class Col:
    TYPE = "type"
    REAL = "real"
    ATTRIBUTE = "attribute"
    NAME = "name"
    DATESTR = "datestr"
    ORIGINAL_PATH = "original_path"
    REL_PATH = "rel_path"


class SurfaceType(str, Enum):
    OBSERVED = "observed"
    SIMULATED = "simulated"


class ProviderImplFile(EnsembleSurfaceProvider):
    def __init__(
        self, provider_id: str, provider_dir: Path, surface_inventory_df: pd.DataFrame
    ) -> None:
        self._provider_id = provider_id
        self._provider_dir = provider_dir
        self._inventory_df = surface_inventory_df

        self._stat_surf_cache = StatSurfCache(self._provider_dir / REL_STAT_CACHE_DIR)

    @staticmethod
    # pylint: disable=too-many-locals
    def write_backing_store(
        storage_dir: Path,
        storage_key: str,
        sim_surfaces: List[SurfaceFileInfo],
        obs_surfaces: List[SurfaceFileInfo],
        avoid_copying_surfaces: bool,
    ) -> None:
        """If avoid_copying_surfaces if True, the specified surfaces will NOT be copied
        into the backing store, but will be referenced from their source locations.
        Note that this is only useful when running in non-portable mode and will fail
        in portable mode.
        """

        timer = PerfTimer()

        do_copy_surfs_into_store = not avoid_copying_surfaces

        # All data for this provider will be stored inside a sub-directory
        # given by the storage key
        provider_dir = storage_dir / storage_key
        LOGGER.debug(f"Writing surface backing store to: {provider_dir}")
        provider_dir.mkdir(parents=True, exist_ok=True)
        (provider_dir / REL_SIM_DIR).mkdir(parents=True, exist_ok=True)
        (provider_dir / REL_OBS_DIR).mkdir(parents=True, exist_ok=True)

        type_arr: List[SurfaceType] = []
        real_arr: List[int] = []
        attribute_arr: List[str] = []
        name_arr: List[str] = []
        datestr_arr: List[str] = []
        rel_path_arr: List[str] = []
        original_path_arr: List[str] = []

        for surfinfo in sim_surfaces:
            type_arr.append(SurfaceType.SIMULATED)
            real_arr.append(surfinfo.real)
            attribute_arr.append(surfinfo.attribute)
            name_arr.append(surfinfo.name)
            datestr_arr.append(surfinfo.datestr if surfinfo.datestr else "")
            original_path_arr.append(surfinfo.path)

            rel_path_in_store = ""
            if do_copy_surfs_into_store:
                rel_path_in_store = _compose_rel_sim_surf_pathstr(
                    real=surfinfo.real,
                    attribute=surfinfo.attribute,
                    name=surfinfo.name,
                    datestr=surfinfo.datestr,
                    extension=Path(surfinfo.path).suffix,
                )
            rel_path_arr.append(rel_path_in_store)

        # We want to strip out observed surfaces without a matching simulated surface
        valid_obs_surfaces = _find_observed_surfaces_corresponding_to_simulated(
            obs_surfaces=obs_surfaces, sim_surfaces=sim_surfaces
        )

        for surfinfo in valid_obs_surfaces:
            type_arr.append(SurfaceType.OBSERVED)
            real_arr.append(-1)
            attribute_arr.append(surfinfo.attribute)
            name_arr.append(surfinfo.name)
            datestr_arr.append(surfinfo.datestr if surfinfo.datestr else "")
            original_path_arr.append(surfinfo.path)

            rel_path_in_store = ""
            if do_copy_surfs_into_store:
                rel_path_in_store = _compose_rel_obs_surf_pathstr(
                    attribute=surfinfo.attribute,
                    name=surfinfo.name,
                    datestr=surfinfo.datestr,
                    extension=Path(surfinfo.path).suffix,
                )
            rel_path_arr.append(rel_path_in_store)

        timer.lap_s()
        if do_copy_surfs_into_store:
            LOGGER.debug(
                f"Copying {len(original_path_arr)} surfaces into backing store..."
            )
            _copy_surfaces_into_provider_dir(
                original_path_arr, rel_path_arr, provider_dir
            )
        et_copy_s = timer.lap_s()

        surface_inventory_df = pd.DataFrame(
            {
                Col.TYPE: type_arr,
                Col.REAL: real_arr,
                Col.ATTRIBUTE: attribute_arr,
                Col.NAME: name_arr,
                Col.DATESTR: datestr_arr,
                Col.REL_PATH: rel_path_arr,
                Col.ORIGINAL_PATH: original_path_arr,
            }
        )

        parquet_file_name = provider_dir / "surface_inventory.parquet"
        surface_inventory_df.to_parquet(path=parquet_file_name)

        if do_copy_surfs_into_store:
            LOGGER.debug(
                f"Wrote surface backing store in: {timer.elapsed_s():.2f}s ("
                f"copy={et_copy_s:.2f}s)"
            )
        else:
            LOGGER.debug(
                f"Wrote surface backing store without copying surfaces in: "
                f"{timer.elapsed_s():.2f}s"
            )

    @staticmethod
    def from_backing_store(
        storage_dir: Path,
        storage_key: str,
    ) -> Optional["ProviderImplFile"]:

        provider_dir = storage_dir / storage_key
        parquet_file_name = provider_dir / "surface_inventory.parquet"

        try:
            surface_inventory_df = pd.read_parquet(path=parquet_file_name)
            return ProviderImplFile(storage_key, provider_dir, surface_inventory_df)
        except FileNotFoundError:
            return None

    def provider_id(self) -> str:
        return self._provider_id

    def attributes(self) -> List[str]:
        return sorted(list(self._inventory_df[Col.ATTRIBUTE].unique()))

    def surface_names_for_attribute(self, surface_attribute: str) -> List[str]:
        return sorted(
            list(
                self._inventory_df.loc[
                    self._inventory_df[Col.ATTRIBUTE] == surface_attribute
                ][Col.NAME].unique()
            )
        )

    def surface_dates_for_attribute(
        self, surface_attribute: str
    ) -> Optional[List[str]]:
        dates = sorted(
            list(
                self._inventory_df.loc[
                    self._inventory_df[Col.ATTRIBUTE] == surface_attribute
                ][Col.DATESTR].unique()
            )
        )
        if len(dates) == 1 and not bool(dates[0]):
            return None

        return dates

    def realizations(self) -> List[int]:
        unique_reals = self._inventory_df[Col.REAL].unique()

        # Sort and strip out any entries with real == -1
        return sorted([r for r in unique_reals if r >= 0])

    def get_surface(
        self,
        address: SurfaceAddress,
    ) -> Optional[xtgeo.RegularSurface]:
        if isinstance(address, StatisticalSurfaceAddress):
            return self._get_or_create_statistical_surface(address)
            # return self._create_statistical_surface(address)
        if isinstance(address, SimulatedSurfaceAddress):
            return self._get_simulated_surface(address)
        if isinstance(address, ObservedSurfaceAddress):
            return self._get_observed_surface(address)

        raise TypeError("Unknown type of surface address")

    def _get_or_create_statistical_surface(
        self, address: StatisticalSurfaceAddress
    ) -> Optional[xtgeo.RegularSurface]:

        timer = PerfTimer()

        surf = self._stat_surf_cache.fetch(address)
        if surf:
            LOGGER.debug(
                f"Fetched statistical surface from cache in: {timer.elapsed_s():.2f}s"
            )
            return surf

        surf = self._create_statistical_surface(address)
        et_create_s = timer.lap_s()

        self._stat_surf_cache.store(address, surf)
        et_write_cache_s = timer.lap_s()

        LOGGER.debug(
            f"Created and wrote statistical surface to cache in: {timer.elapsed_s():.2f}s ("
            f"create={et_create_s:.2f}s, store={et_write_cache_s:.2f}s), "
            f"[stat={address.statistic}, "
            f"attr={address.attribute}, name={address.name}, date={address.datestr}]"
        )

        return surf

    def _create_statistical_surface(
        self, address: StatisticalSurfaceAddress
    ) -> Optional[xtgeo.RegularSurface]:
        surf_fns: List[str] = self._locate_simulated_surfaces(
            attribute=address.attribute,
            name=address.name,
            datestr=address.datestr if address.datestr is not None else "",
            realizations=address.realizations,
        )

        if len(surf_fns) == 0:
            LOGGER.warning(f"No input surfaces found for statistical surface {address}")
            return None

        timer = PerfTimer()

        surfaces = xtgeo.Surfaces(surf_fns)
        et_load_s = timer.lap_s()

        surf_count = len(surfaces.surfaces)
        if surf_count == 0:
            LOGGER.warning(
                f"Could not load input surfaces for statistical surface {address}"
            )
            return None

        # print("########################################################")
        # first_surf = surfaces.surfaces[0]
        # for surf in surfaces.surfaces:
        #     print(
        #         surf.dimensions,
        #         surf.xinc,
        #         surf.yinc,
        #         surf.xori,
        #         surf.yori,
        #         surf.rotation,
        #         surf.filesrc,
        #     )
        # print("########################################################")

        # Suppress numpy warnings when surfaces have undefined z-values
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", "All-NaN slice encountered")
            warnings.filterwarnings("ignore", "Mean of empty slice")
            warnings.filterwarnings("ignore", "Degrees of freedom <= 0 for slice")

            stat_surface = _calc_statistic_across_surfaces(address.statistic, surfaces)
        et_calc_s = timer.lap_s()

        LOGGER.debug(
            f"Created statistical surface in: {timer.elapsed_s():.2f}s ("
            f"load={et_load_s:.2f}s, calc={et_calc_s:.2f}s), "
            f"[#surfaces={surf_count}, stat={address.statistic}, "
            f"attr={address.attribute}, name={address.name}, date={address.datestr}]"
        )

        return stat_surface

    def _get_simulated_surface(
        self, address: SimulatedSurfaceAddress
    ) -> Optional[xtgeo.RegularSurface]:
        """Returns a Xtgeo surface instance of a single realization surface"""

        timer = PerfTimer()

        surf_fns: List[str] = self._locate_simulated_surfaces(
            attribute=address.attribute,
            name=address.name,
            datestr=address.datestr if address.datestr is not None else "",
            realizations=[address.realization],
        )

        if len(surf_fns) == 0:
            LOGGER.warning(f"No simulated surface found for {address}")
            return None
        if len(surf_fns) > 1:
            LOGGER.warning(
                f"Multiple simulated surfaces found for: {address}"
                "Returning first surface."
            )

        surf = xtgeo.surface_from_file(surf_fns[0])

        LOGGER.debug(f"Loaded simulated surface in: {timer.elapsed_s():.2f}s")

        return surf

    def _get_observed_surface(
        self, address: ObservedSurfaceAddress
    ) -> Optional[xtgeo.RegularSurface]:
        """Returns a Xtgeo surface instance for an observed surface"""

        timer = PerfTimer()

        surf_fns: List[str] = self._locate_observed_surfaces(
            attribute=address.attribute,
            name=address.name,
            datestr=address.datestr if address.datestr is not None else "",
        )

        if len(surf_fns) == 0:
            LOGGER.warning(f"No observed surface found for {address}")
            return None
        if len(surf_fns) > 1:
            LOGGER.warning(
                f"Multiple observed surfaces found for: {address}"
                "Returning first surface."
            )

        surf = xtgeo.surface_from_file(surf_fns[0])

        LOGGER.debug(f"Loaded simulated surface in: {timer.elapsed_s():.2f}s")

        return surf

    def _locate_simulated_surfaces(
        self, attribute: str, name: str, datestr: str, realizations: List[int]
    ) -> List[str]:
        """Returns list of file names matching the specified filter criteria"""
        df = self._inventory_df.loc[
            self._inventory_df[Col.TYPE] == SurfaceType.SIMULATED
        ]

        df = df.loc[
            (df[Col.ATTRIBUTE] == attribute)
            & (df[Col.NAME] == name)
            & (df[Col.DATESTR] == datestr)
            & (df[Col.REAL].isin(realizations))
        ]

        df = df[[Col.REL_PATH, Col.ORIGINAL_PATH]]

        # Return file name within backing store if the surface was copied there,
        # otherwise return the original source file name
        fn_list: List[str] = []
        for _index, row in df.iterrows():
            if row[Col.REL_PATH]:
                fn_list.append(self._provider_dir / row[Col.REL_PATH])
            else:
                fn_list.append(row[Col.ORIGINAL_PATH])

        return fn_list

    def _locate_observed_surfaces(
        self, attribute: str, name: str, datestr: str
    ) -> List[str]:
        """Returns file names of observed surfaces matching the criteria"""
        df = self._inventory_df.loc[
            self._inventory_df[Col.TYPE] == SurfaceType.OBSERVED
        ]

        df = df.loc[
            (df[Col.ATTRIBUTE] == attribute)
            & (df[Col.NAME] == name)
            & (df[Col.DATESTR] == datestr)
        ]

        df = df[[Col.REL_PATH, Col.ORIGINAL_PATH]]

        # Return file name within backing store if the surface was copied there,
        # otherwise return the original source file name
        fn_list: List[str] = []
        for _index, row in df.iterrows():
            if row[Col.REL_PATH]:
                fn_list.append(self._provider_dir / row[Col.REL_PATH])
            else:
                fn_list.append(row[Col.ORIGINAL_PATH])

        return fn_list


def _find_observed_surfaces_corresponding_to_simulated(
    obs_surfaces: List[SurfaceFileInfo], sim_surfaces: List[SurfaceFileInfo]
) -> List[SurfaceFileInfo]:
    """Returns only the observed surfaces that have a matching simulated surface"""

    unique_sim_surf_ids: Set[str] = set()
    for surfinfo in sim_surfaces:
        surf_id = f"{surfinfo.name}_{surfinfo.attribute}_{surfinfo.datestr}"
        unique_sim_surf_ids.add(surf_id)

    valid_obs_surfaces: List[SurfaceFileInfo] = []
    for surfinfo in obs_surfaces:
        surf_id = f"{surfinfo.name}_{surfinfo.attribute}_{surfinfo.datestr}"
        if surf_id in unique_sim_surf_ids:
            valid_obs_surfaces.append(surfinfo)
        else:
            LOGGER.debug(
                f"Discarding observed surface without matching simulation surface {surfinfo.path}"
            )

    return valid_obs_surfaces


def _copy_surfaces_into_provider_dir(
    original_path_arr: List[str],
    rel_path_arr: List[str],
    provider_dir: Path,
) -> None:
    for src_path, dst_rel_path in zip(original_path_arr, rel_path_arr):
        # LOGGER.debug(f"copying surface from: {src_path}")
        shutil.copyfile(src_path, provider_dir / dst_rel_path)

    # full_dst_path_arr = [storage_dir / dst_rel_path for dst_rel_path in store_path_arr]
    # with ProcessPoolExecutor() as executor:
    #     executor.map(shutil.copyfile, original_path_arr, full_dst_path_arr)


def _compose_rel_sim_surf_pathstr(
    real: int,
    attribute: str,
    name: str,
    datestr: Optional[str],
    extension: str,
) -> str:
    """Compose path to simulated surface file, relative to provider's directory"""
    if datestr:
        fname = f"{real}--{name}--{attribute}--{datestr}{extension}"
    else:
        fname = f"{real}--{name}--{attribute}{extension}"
    return str(Path(REL_SIM_DIR) / fname)


def _compose_rel_obs_surf_pathstr(
    attribute: str,
    name: str,
    datestr: Optional[str],
    extension: str,
) -> str:
    """Compose path to observed surface file, relative to provider's directory"""
    if datestr:
        fname = f"{name}--{attribute}--{datestr}{extension}"
    else:
        fname = f"{name}--{attribute}{extension}"
    return str(Path(REL_OBS_DIR) / fname)


def _calc_statistic_across_surfaces(
    statistic: SurfaceStatistic, surfaces: xtgeo.Surfaces
) -> xtgeo.RegularSurface:
    """Calculates a statistical surface from a list of Xtgeo surface instances"""

    stat_surf: xtgeo.RegularSurface

    if statistic == SurfaceStatistic.MEAN:
        stat_surf = surfaces.apply(np.mean, axis=0)
    elif statistic == SurfaceStatistic.STDDEV:
        stat_surf = surfaces.apply(np.std, axis=0)
    elif statistic == SurfaceStatistic.MINIMUM:
        stat_surf = surfaces.apply(np.min, axis=0)
    elif statistic == SurfaceStatistic.MAXIMUM:
        stat_surf = surfaces.apply(np.max, axis=0)
    elif statistic == SurfaceStatistic.P10:
        stat_surf = surfaces.apply(np.percentile, 10, axis=0)
    elif statistic == SurfaceStatistic.P90:
        stat_surf = surfaces.apply(np.percentile, 90, axis=0)

    return stat_surf



from typing import Callable, Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
from dash import Dash, Input, Output, State
from dash.exceptions import PreventUpdate

from webviz_subsurface._models import SurfaceSetModel, WellSetModel


def update_uncertainty_table(
    app: Dash,
    get_uuid: Callable,
    surface_set_models: Dict[str, SurfaceSetModel],
    well_set_model: WellSetModel,
) -> None:
    @app.callback(
        Output({"id": get_uuid("uncertainty-table"), "element": "table"}, "data"),
        Output({"id": get_uuid("uncertainty-table"), "element": "label"}, "children"),
        Input(
            {"id": get_uuid("uncertainty-table"), "element": "apply-button"}, "n_clicks"
        ),
        State({"id": get_uuid("intersection-data"), "element": "well"}, "value"),
        State(
            {"id": get_uuid("intersection-data"), "element": "surface_attribute"},
            "value",
        ),
        State(
            {"id": get_uuid("intersection-data"), "element": "surface_names"}, "value"
        ),
        State({"id": get_uuid("intersection-data"), "element": "ensembles"}, "value"),
        State(get_uuid("realization-store"), "data"),
    )
    # pylint: disable=too-many-locals
    def _update_uncertainty_table(
        apply_btn: Optional[int],
        wellname: str,
        surface_attribute: str,
        surface_names: List[str],
        ensembles: List[str],
        realizations: List[int],
    ) -> Tuple[List, str]:
        if apply_btn is None:
            raise PreventUpdate
        dframes = []

        well = well_set_model.get_well(wellname)
        realizations = [int(real) for real in realizations]
        for ensemble in ensembles:
            surfset = surface_set_models[ensemble]
            for surfacename in surface_names:
                for calculation in ["Mean", "Min", "Max"]:
                    surface = surfset.calculate_statistical_surface(
                        name=surfacename,
                        attribute=surface_attribute,
                        calculation=calculation,
                        realizations=realizations,
                    )

                    with np.errstate(invalid="ignore"):
                        surface_picks = well.get_surface_picks(surface)
                        if surface_picks is None:
                            continue
                        surface_df = surface_picks.dataframe.drop(
                            columns=["X_UTME", "Y_UTMN", "DIRECTION", "WELLNAME"]
                        )
                        tvds = (
                            surface_df["Z_TVDSS"]
                            .apply(lambda x: np.round(x, 1))
                            .tolist()
                        )

                        surface_df.rename({well.mdlogname: "MD"}, axis=1, inplace=True)
                        # Do not calculate MD if Well tvd is truncated
                        mds = (
                            surface_df["MD"].apply(lambda x: np.round(x, 1)).tolist()
                            if not well_set_model.is_truncated
                            else []
                        )

                        surface_df["Pick no"] = surface_df.index + 1
                        pick_nos = surface_df["Pick no"].tolist()

                        data = {
                            "TVD": " / ".join(str(x) for x in tvds),
                            "MD": " / ".join(str(x) for x in mds),
                            "Pick no": " / ".join(str(x) for x in pick_nos),
                            "Surface name": surfacename,
                            "Calculation": calculation,
                            "Ensemble": ensemble,
                        }
                        surface_df = pd.DataFrame([data])
                        dframes.append(surface_df)

        dframe = (
            pd.concat(dframes)
            if dframes
            else pd.DataFrame(
                [
                    {
                        "TVD": "-",
                        "MD": "-",
                        "Pick no": "-",
                        "Surface name": "-",
                        "Calculation": "-",
                        "Ensemble": "-",
                    }
                ]
            )
        )
        return dframe.to_dict("records"), f"Statistics for well {wellname}"



from typing import Callable, Dict, List, Optional

from dash import ALL, ClientsideFunction, Dash, Input, Output, callback_context
from dash.exceptions import PreventUpdate
from plotly.graph_objects import Figure


def update_figure_clientside(app: Dash, get_uuid: Callable) -> None:
    @app.callback(
        Output(
            {"id": get_uuid("clientside"), "plotly_attribute": "update_layout"}, "data"
        ),
        Input({"id": get_uuid("plotly_layout"), "layout_attribute": ALL}, "value"),
        Input(get_uuid("plotly_uirevision"), "value"),
    )
    def _update_layout(layout_attributes: Optional[List], uirevision: str) -> Dict:
        """Store plotly layout options from user selections in a dcc.Store"""
        if layout_attributes is None:
            return {}
        layout = {}
        for ctx in callback_context.inputs_list[0]:
            layout[ctx["id"]["layout_attribute"]] = ctx.get("value", None)
        layout["uirevision"] = str(uirevision) if uirevision else None
        return layout

    @app.callback(
        Output(
            {"id": get_uuid("clientside"), "plotly_attribute": "plotly_layout"}, "data"
        ),
        Input(
            {"id": get_uuid("clientside"), "plotly_attribute": "initial_layout"}, "data"
        ),
        Input(
            {"id": get_uuid("clientside"), "plotly_attribute": "update_layout"}, "data"
        ),
    )
    def _update_plot_layout(initial_layout: dict, update_layout: dict) -> Dict:
        if initial_layout is None:
            raise PreventUpdate
        fig = Figure({"layout": initial_layout})
        if update_layout is not None:
            fig.update_layout(update_layout)
        return fig["layout"]

    app.clientside_callback(
        ClientsideFunction(namespace="clientside", function_name="set_dcc_figure"),
        Output(get_uuid("graph"), "figure"),
        Input(
            {"id": get_uuid("clientside"), "plotly_attribute": "plotly_layout"}, "data"
        ),
        Input(
            {"id": get_uuid("clientside"), "plotly_attribute": "plotly_data"}, "data"
        ),
    )



import json

from dash import Input, Output, State, html
from webviz_config import WebvizSettings

# pylint: disable=no-name-in-module
from webviz_config.plugins import StructuralUncertainty
from webviz_config.themes import default_theme

# pylint: enable=no-name-in-module


def stringify_object_id(uuid) -> str:
    """Object ids must be sorted and converted to
    css strings to be recognized as dom elements"""
    sorted_uuid_obj = json.loads(
        json.dumps(
            uuid,
            sort_keys=True,
            separators=(",", ":"),
        )
    )
    string = ["{"]
    for idx, (key, value) in enumerate(sorted_uuid_obj.items()):
        string.append(f'\\"{key}\\"\\:\\"{value}\\"\\')
        if idx == len(sorted_uuid_obj) - 1:
            string.append("}")
        else:
            string.append(",")
    return ("").join(string)


# pylint: disable=too-many-locals
def test_default_configuration(dash_duo, app, testdata_folder) -> None:
    webviz_settings = WebvizSettings(
        shared_settings={
            "scratch_ensembles": {
                "iter-0": str(testdata_folder / "01_drogon_ahm/realization-*/iter-0")
            }
        },
        theme=default_theme,
    )
    plugin = StructuralUncertainty(
        app,
        webviz_settings,
        ensembles=["iter-0"],
        surface_attributes=["ds_extract_postprocess"],
        surface_name_filter=[
            "topvolon",
            "toptherys",
            "topvolantis",
            "basevolantis",
        ],
        wellsuffix=".rmswell",
        wellfolder=testdata_folder / "observed_data" / "wells",
    )

    app.layout = plugin.layout
    dash_duo.start_server(app)

    intersection_data_id = plugin.uuid("intersection-data")
    dialog_id = plugin.uuid("dialog")
    # Check some initialization
    # Check dropdowns
    for element, return_val in zip(
        ["well", "surface_attribute"], ["55_33-1", "ds_extract_postprocess"]
    ):
        uuid = stringify_object_id(
            uuid={"element": element, "id": intersection_data_id}
        )
        assert dash_duo.wait_for_element(f"#\\{uuid} .Select-value").text == return_val

    # Check Selects
    for element, return_val in zip(
        ["surface_names"],
        [["topvolon", "toptherys", "topvolantis", "basevolantis"]],
    ):
        uuid = stringify_object_id(
            uuid={"element": element, "id": intersection_data_id}
        )
        assert (
            dash_duo.wait_for_element(f"#\\{uuid} select").text.splitlines()
            == return_val
        )

    # Check Calculation checkbox
    uuid = stringify_object_id(
        uuid={"element": "calculation", "id": intersection_data_id}
    )
    calculation_element = dash_duo.driver.find_elements_by_css_selector(
        f"#\\{uuid} > label > input"
    )
    assert len(calculation_element) == len(
        ["Min", "Max", "Mean", "Realizations", "Uncertainty envelope"]
    )
    for checkbox, selected in zip(
        calculation_element,
        ["true", "true", "true", None, None],
    ):
        assert checkbox.get_attribute("selected") == selected

    # Check realizations
    real_filter_btn_uuid = stringify_object_id(
        {
            "id": dialog_id,
            "dialog_id": "realization-filter",
            "element": "button-open",
        }
    )
    real_uuid = stringify_object_id(
        uuid={"element": "realizations", "id": intersection_data_id}
    )

    ### Open realization filter and check realizations
    dash_duo.wait_for_element_by_id(real_filter_btn_uuid).click()
    real_selector = dash_duo.wait_for_element_by_id(real_uuid)
    assert real_selector.text.splitlines() == ["0", "1"]

    assert dash_duo.get_logs() == [], "browser console should contain no error"


def test_full_configuration(dash_duo, app, testdata_folder) -> None:
    webviz_settings = WebvizSettings(
        shared_settings={
            "scratch_ensembles": {
                "iter-0": str(testdata_folder / "01_drogon_ahm/realization-*/iter-0"),
            }
        },
        theme=default_theme,
    )
    plugin = StructuralUncertainty(
        app,
        webviz_settings,
        ensembles=["iter-0"],
        surface_attributes=["ds_extract_postprocess"],
        surface_name_filter=["topvolon", "toptherys", "topvolantis", "basevolantis"],
        wellfolder=testdata_folder / "observed_data" / "wells",
        wellsuffix=".rmswell",
        zonelog="Zone",
        initial_settings={
            "intersection_data": {
                "surface_names": ["topvolon", "toptherys", "topvolantis"],
                "surface_attribute": "ds_extract_postprocess",
                "ensembles": [
                    "iter-0",
                ],
                "calculation": ["Mean", "Min", "Max"],
                # - Uncertainty envelope
                "well": "55_33-1",
                "realizations": [0, 1],
                "colors": {
                    "topvolon": {"iter-0": "#2C82C9"},
                    "toptherys": {
                        "iter-0": "#512E34",
                    },
                    "topvolantis": {
                        "iter-0": "#EEE657",
                    },
                },
            },
            "intersection_layout": {
                "yaxis": {
                    "range": [1700, 1550],
                    "title": "True vertical depth [m]",
                },
                "xaxis": {"title": "Lateral distance [m]"},
            },
        },
    )

    app.layout = plugin.layout

    # Injecting a div that will be updated when the plot data stores are
    # changed. Since the plot data are stored in LocalStorage and Selenium
    # has no functionality to wait for LocalStorage to equal some value we
    # instead populate this injected div with some data before we check the content
    # of Localstorage.
    @app.callback(
        Output(plugin.uuid("layout"), "children"),
        Input(plugin.uuid("intersection-graph-layout"), "data"),
        State(plugin.uuid("layout"), "children"),
    )
    def _add_or_update_div(data, children):
        plot_is_updated = html.Div(
            id=plugin.uuid("plot_is_updated"), children=data.get("title")
        )
        if len(children) == 6:
            children[5] = plot_is_updated
        else:
            children.append(plot_is_updated)

        return children

    dash_duo.start_server(app)

    intersection_data_id = plugin.uuid("intersection-data")

    # Check some initialization
    # Check dropdowns
    for element, return_val in zip(
        ["well", "surface_attribute"], ["55_33-1", "ds_extract_postprocess"]
    ):
        uuid = stringify_object_id(
            uuid={"element": element, "id": intersection_data_id}
        )
        assert dash_duo.wait_for_text_to_equal(f"#\\{uuid} .Select-value", return_val)

    # Wait for the callbacks to execute
    dash_duo.wait_for_text_to_equal(
        f'#{plugin.uuid("plot_is_updated")}',
        "Intersection along well: 55_33-1",
        timeout=30,
    )

    # Check that graph data is stored
    graph_data = dash_duo.get_session_storage(plugin.uuid("intersection-graph-data"))
    assert len(graph_data) == 14
    graph_layout = dash_duo.get_session_storage(
        plugin.uuid("intersection-graph-layout")
    )
    assert isinstance(graph_layout, dict)
    assert graph_layout.get("title") == "Intersection along well: 55_33-1"

    ### Change well and check graph
    well_uuid = stringify_object_id(
        uuid={"element": "well", "id": intersection_data_id}
    )

    apply_btn = dash_duo.wait_for_element_by_id(
        plugin.uuid("apply-intersection-data-selections")
    )
    well_dropdown = dash_duo.wait_for_element_by_id(well_uuid)
    dash_duo.select_dcc_dropdown(well_dropdown, value="55_33-2")
    apply_btn.click()

    # dash_duo.wait_for_text_to_equal(
    #     f'#{plugin.uuid("plot_is_updated")}',
    #     "Intersection along well: 55_33-1",
    #     timeout=100,
    # )
    graph_layout = dash_duo.get_session_storage(
        plugin.uuid("intersection-graph-layout")
    )
    # assert graph_layout.get("title") == "Intersection along well: 55_33-2"
